import asyncio
from abc import ABC, abstractmethod
from asyncio import AbstractEventLoop
from json import dumps, loads
from os import _exit
from typing import Callable, Dict

import zmq
import zmq.asyncio
from pydantic import BaseModel

from backtesterrb30.libs.interfaces.utils.config import Config
from backtesterrb30.libs.utils.list_of_services import SERVICES
from backtesterrb30.libs.utils.service import Service


class ZMQ(Service, ABC):
    __is_running: bool
    __is_active: bool

    __commands: Dict[str, Callable]

    # override
    def __init__(self, config: Config, logger=print):
        super().__init__(config, logger)
        self.config = config
        self.__is_running = False
        self.__is_active = False

        self.__pub = None
        self.__subs = []

        self.__commands = {
            "start": self._start,
            "pause": self._pause,
            "stop": self._stop,
        }

    # override
    def run(self):
        self.__init_sockets(self.config)
        self.__is_running = True
        self.__is_active = True
        super().run()

    # override
    def _send(self, service: SERVICES, msg: str, *args):
        if self.__is_active:
            data = [service.value.encode("utf-8"), msg.encode("utf-8")]
            for arg in args:
                if isinstance(arg, BaseModel):
                    arg = arg.dict()
                arg = dumps(arg, default=str)
                data.append(arg.encode("utf-8"))
            self.__pub.send_multipart(data)

    @abstractmethod
    def _handle_zmq_message(self, msg: str):
        pass

    @abstractmethod
    def _asyncio_loop(self, loop: asyncio.AbstractEventLoop):
        pass

    def _loop(self):
        loop = asyncio.get_event_loop()
        self._asyncio_loop(loop)
        loop.run_forever()
        loop.close()

    def _register(self, command: str, func: Callable):
        self.__commands[command] = func

    def _create_listeners(self, loop: AbstractEventLoop):
        self._log("Start main loop")
        for sub in self.__subs:
            loop.create_task(self.__listen_zmq(sub))

    async def __listen_zmq(self, sub):
        while self.__is_running:
            if self.__is_active:
                data = await sub.recv_multipart()
                if data:
                    await self.__handle(data)
                await asyncio.sleep(0.00001)
            else:
                await asyncio.sleep(0.01)
        self.__deinit()

    def __init_sockets(self, config: Config):
        sub_context = zmq.asyncio.Context()
        pub_context = zmq.Context()
        for sub in config.sub:
            s = sub_socket(
                sub_context, "tcp://" + config.ip + ":" + str(sub.port), sub.topic
            )
            self.__subs.append(s)
        self.__pub = pub_socket(pub_context, "tcp://*:" + str(config.pub.port))

    async def __handle(self, data):
        data = [x.decode("utf-8") for x in data]
        if len(data) >= 2:
            topic, cmd, *args = data
            func = self.__commands.get(cmd)
            if func:
                # self._log(f'Receive "{cmd}" command')
                if args:
                    args_loaded = []
                    for arg in args:
                        args_loaded.append(loads(arg))
                    await func(*args_loaded)
                else:
                    await func()
            else:
                self._log(f"Command '{cmd}' not registered")
        self._handle_zmq_message(data)

    def __deinit(self):
        self.__pub.close()

    def _stop(self):
        self._log("Service stoped")
        self.__is_running = False
        _exit(1)

    def _start(self):
        self._log("Service started")
        self.__is_active = True

    def _pause(self):
        self._log("Service paused")
        self.__is_active = False


def sub_socket(context: zmq.asyncio.Context(), url: str, topic: str):
    sub = context.socket(zmq.SUB)
    sub.connect(url)
    sub.subscribe(topic)
    return sub


def pub_socket(context: zmq.Context(), url: str):
    pub = context.socket(zmq.PUB)
    pub.bind(url)
    return pub



from datetime import datetime
from enum import Enum

# from symbol import parameters
from pydantic import BaseModel

# from backtesterrb30.libs.utils.historical_sources import HISTORICAL_SOURCES, HISTORICAL_INTERVALS_UNION
from backtesterrb30.libs.utils.timestamps import datetime_to_timestamp


class InstrumentFile(BaseModel):
    source: str
    instrument: str
    interval: str
    time_start: int
    time_stop: int

    @property
    def identifier(self):
        return self.source + "_" + self.instrument

    @classmethod
    def from_filename(cls, filename: str):
        if not filename or filename == "":
            raise Exception("Empty file name name")
        if filename[-4:] != ".csv":
            raise Exception("Bad file extension")
        filename = filename[:-4]
        source, instrument, interval, time_start, time_stop = tuple(
            filename.split("__")
        )
        if None in (source, instrument, interval, time_start, time_stop) or "" in (
            source,
            instrument,
            interval,
            time_start,
            time_stop,
        ):
            raise Exception("Bad file name string provided")
        dict_instrument_file = {
            "source": source,
            "instrument": instrument,
            "interval": interval,
            "time_start": int(time_start),
            "time_stop": int(time_stop),
        }
        return cls(**dict_instrument_file)

    @classmethod
    def from_params(
        cls,
        source: str,
        instrument: str,
        interval: Enum,
        time_start: datetime,
        time_stop: datetime,
    ):
        dict_instrument_file = {
            "source": source,
            "instrument": instrument,
            "interval": interval.value,
            "time_start": datetime_to_timestamp(time_start),
            "time_stop": datetime_to_timestamp(time_stop),
        }
        return cls(**dict_instrument_file)

    def to_filename(self):
        return (
            self.source
            + "__"
            + self.instrument
            + "__"
            + self.interval
            + "__"
            + str(self.time_start)
            + "__"
            + str(self.time_stop)
            + ".csv"
        )

    def __str__(self):
        return self.to_filename()

    def toJSON(self):
        return self.to_filename()

    # def get_file_name() -> str:



from backtesterrb30.libs.utils.run_service import run_service
from backtesterrb30.python_backtester.python_backtester import Backtester

microservice_name = "python_backtester"

run_service(microservice_name, Backtester)



import asyncio
import time as tm
from os import listdir, makedirs, path
from os.path import isfile, join
from typing import List

from appdirs import user_cache_dir

from backtesterrb30.historical_data_feeds.functions import (
    get_instrument_files,
    load_data_frame_ticks_2,
)
from backtesterrb30.historical_data_feeds.historical_downloader import (
    HistoricalDownloader,
)
from backtesterrb30.libs.communication_broker.broker_base import BrokerBase
from backtesterrb30.libs.data_sources.data_sources_list import HISTORICAL_SOURCES
from backtesterrb30.libs.interfaces.historical_data_feeds.instrument_file import (
    InstrumentFile,
)
from backtesterrb30.libs.interfaces.python_backtester.data_start import DataStart
from backtesterrb30.libs.interfaces.utils.config import Config
from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema
from backtesterrb30.libs.utils.list_of_services import SERVICES, SERVICES_ARRAY
from backtesterrb30.libs.utils.service import Service


class HistoricalDataFeeds(Service):
    downloaded_data_path = user_cache_dir("rb30_cache")
    _broker: BrokerBase

    def __init__(
        self, config: Config, data_schema: DataSchema, loop=None, logger=print
    ):
        super().__init__(config, logger)
        self.config: Config = config
        self.__loop = loop
        self.__custom_event_loop = False
        if self.__loop is None:
            self.__loop = asyncio.get_event_loop()
            self.__custom_event_loop = True
        self.__data_schema: DataSchema = data_schema
        self.__columns = ["timestamp"] + [c.symbol for c in self.__data_schema.data]
        self.__historical_sources_array = [
            i for i in dir(HISTORICAL_SOURCES) if not i.startswith("__")
        ]
        self.__data_sources = {}
        self.__sending_locked = False
        self.__start_time = 0
        self.__data_to_download_2 = None
        self.__engine_ready = False
        self.__historical_downloader = HistoricalDownloader(
            self.__data_schema,
            self.__historical_sources_array,
            self.downloaded_data_path,
            self._log,
        )

    # # override
    # def _asyncio_loop(self, loop: asyncio.AbstractEventLoop):
    #     self._create_listeners(loop)
    #     self.__validate_downloaded_data_folder()
    #     if self.config.backtest:
    #         self.__data_to_download_2 = self.__historical_downloader.run(loop)
    #         data_parts = self.__prepare_loading_data_structure_2()
    #         loop.create_task(self.__historical_data_loop_ticks(data_parts))
    #     else:
    #         self._log('backtest is off')

    def _loop(self):
        # self._broker.run()
        self._broker.create_listeners(self.__loop)
        self.__validate_downloaded_data_folder()
        if self.config.backtest:
            self.__data_to_download_2 = self.__historical_downloader.run(self.__loop)
            data_parts = self.__prepare_loading_data_structure_2()
            self.__loop.create_task(self.__historical_data_loop_ticks(data_parts))
        else:
            self._log("backtest is off")
        if self.__custom_event_loop:
            self.__loop.run_forever()
            self.__loop.close()

    # def _send(self, service: SERVICES, msg: str, *args):
    #     self._broker.send(service, msg, *args)

    def _configure(self):
        super()._configure()
        # register commands
        self._broker.register(
            "unlock_historical_sending", self.__unlock_historical_sending_event
        )
        self._broker.register(
            "engine_ready_response", self.__engine_ready_response_event
        )

    # # override
    # def _handle_zmq_message(self, message):
    #     pass

    async def __send_start_params(self):
        self.__start_time = tm.time()
        file_names_grouped = []
        for symbol in self.__data_schema.data:
            file_names_grouped.append(
                {
                    "symbol": symbol.symbol,
                    "source": symbol.historical_data_source,
                    "files": get_instrument_files(symbol),
                }
            )
        start_params = {
            "file_names": file_names_grouped,
            "start_time": self.__start_time,
        }
        await self._broker.send(
            SERVICES.python_backtester, "data_start", DataStart(**start_params)
        )

    async def __historical_data_loop_ticks(self, data_parts: dict):
        # waiting for zero mq ports starts up
        await asyncio.sleep(0.5)
        while True:
            # if self.__data_downloaded(self.__data_to_download):
            await self._broker.send(
                SERVICES.python_engine,
                "engine_ready",
                SERVICES.historical_data_feeds.value,
            )
            if (
                self.__validate_data_downloaded(self.__data_to_download_2)
                and self.__engine_ready
            ):
                self._log("All data has been downloaded")

                sending_counter = 0
                self._log("Starting data loop")
                await self.__send_start_params()
                # return
                last_row = []
                for _, one_year_array in data_parts.items():
                    self._log("Synchronizing part of data")
                    data_part = load_data_frame_ticks_2(
                        self.__data_schema,
                        self.__columns,
                        self.downloaded_data_path,
                        last_row,
                        one_year_array,
                    )
                    for row in data_part:
                        last_row = row
                        await self._broker.send(
                            SERVICES.python_engine, "data_feed", list(last_row)
                        )
                        sending_counter += 1
                        if sending_counter % 1000 == 0:
                            self.__sending_locked = True
                            await self._broker.send(
                                SERVICES.python_engine, "historical_sending_locked"
                            )
                            while self.__sending_locked:
                                await asyncio.sleep(0.01)

                await self._broker.send(SERVICES.python_engine, "data_finish")
                break
            await asyncio.sleep(1)

    def __validate_data_downloaded(self, files_to_download: List[InstrumentFile]):
        if files_to_download is None:
            return False
        files_in_directory = [
            f
            for f in listdir(self.downloaded_data_path)
            if isfile(join(self.downloaded_data_path, f))
        ]
        data_to_download = list(
            set([f.to_filename() for f in files_to_download]) - set(files_in_directory)
        )
        return data_to_download == []

    def __validate_downloaded_data_folder(self):
        if not path.exists(self.downloaded_data_path):
            makedirs(self.downloaded_data_path)

    def __prepare_loading_data_structure_2(self) -> dict:
        files_collection = {}
        for symbol in self.__data_schema.data:
            files: List[InstrumentFile] = get_instrument_files(symbol)
            for f in files:
                if f.time_stop not in files_collection:
                    files_collection[f.time_stop]: List[InstrumentFile] = []
                files_collection[f.time_stop].append(f)
        return dict(sorted(files_collection.items()))

    async def __stop_all_services(self):
        for service in SERVICES_ARRAY:
            if service != self.name:
                await self._broker.send(getattr(SERVICES, service), "stop")
        self._broker.stop()

    # COMMANDS

    async def __unlock_historical_sending_event(self):
        self.__sending_locked = False

    async def __engine_ready_response_event(self):
        self.__engine_ready = True

    # def __map_raw_to_instruments(self, raw: list, instruments: list):
    #     last_raw_obj = {}
    #     if len(raw) != len(instruments):
    #         raise Exception('Error in map_raw_to_instruments. Lengths of raw and list of instruments are not equal')
    #     for value, instrument in zip(raw, instruments):
    #         last_raw_obj[instrument] = value
    #     return last_raw_obj

    # def __prepare_dataframes_to_synchronize_2(self, downloaded_data_path: str,
    #                 last_row: list, files_array: List[InstrumentFile]) -> List[dict]:
    #     list_of_dfs = []
    #     for data_element in self.__data_schema.data:
    #         columns = ['timestamp', data_element.symbol]
    #         file_name = 'none'
    #         actual_raw = [0,0]
    #         prev_raw = [0,0]
    #         for element in files_array:
    #             if data_element.symbol == element.instrument:
    #                 file_name = element.to_filename()
    #         if file_name == 'none':
    #             # No file in this period for this instrument. Set empty dataframe.
    #             df = pd.DataFrame([], columns=columns)
    #         else:
    #             # File exists. Load dataframe.
    #             df = pd.read_csv(join(downloaded_data_path, file_name), index_col=None, header=None, names=columns)
    #             # append last raw it if exists
    #             if last_row != []:
    #                 # self._log('appending last row')
    #                 last_raw_mapped = self.__map_raw_to_instruments(last_row, self.__columns)
    #                 prev_raw[0] = last_raw_mapped["timestamp"]
    #                 prev_raw[1] = last_raw_mapped[data_element.symbol]
    #         obj = {
    #             "trigger_feed": data_element.trigger_feed,
    #             "rows_iterator": df.iterrows(),
    #             "actual_raw": actual_raw,
    #             "prev_raw": prev_raw,
    #             "consumed": False
    #         }
    #         #prepare to load:
    #         if obj['actual_raw'][0] == 0:
    #             try:
    #                 i, v = next(obj['rows_iterator'])
    #                 obj['actual_raw'] = list(v)
    #             except StopIteration:
    #                 obj['consumed'] = True
    #         list_of_dfs.append(obj)
    #     return list_of_dfs

    # def __load_data_frame_ticks_2(self, downloaded_data_path: str, last_row: list, files_array: List[InstrumentFile]) -> List[list]:
    #     """
    #     Function is geting files array from one period that are going to be loaded.
    #     Function returns synchronized data in list of lists which are ready to send to engine.
    #     """
    #     list_of_dfs = self.__prepare_dataframes_to_synchronize_2(downloaded_data_path, last_row, files_array)
    #     rows = synchronize_dataframes(list_of_dfs, last_row)
    #     return rows

    # def __get_data_to_download(self, data_schema: DataSchema):
    #     for data_symbol in data_schema.data:
    #         data_symbol.additional_properties['files_to_download']: List[InstrumentFile] = []
    #         data_symbol.additional_properties['files_to_download'] = self.__check_symbol_data_exists(data_symbol)
    #         if self.__data_to_download_2 is None:
    #             self.__data_to_download_2 = []
    #         self.__data_to_download_2 = self.__data_to_download_2 + data_symbol.additional_properties['files_to_download']

    # def __create_downloading_clients(self, historical_sources_array: list, data_schema: DataSchema, data_sources: dict):
    #     for source in historical_sources_array:
    #         if source in [data.historical_data_source for data in data_schema.data if \
    #                 data.additional_properties['files_to_download'] != []]:
    #             data_sources[source]: DataSource = getattr(HISTORICAL_SOURCES, source)(self._log)

    # def __validate_symbols_to_download(self,data_schema: DataSchema, loop: asyncio.AbstractEventLoop):
    #     self._log('validating symbols before download')
    #     for symbol in data_schema.data:
    #         if symbol.additional_properties['files_to_download'] != []:
    #             data_source_client: DataSource = self.__get_data_source_client(symbol.historical_data_source)
    #             res = loop.run_until_complete(data_source_client.validate_instrument(symbol))
    #             if not res:
    #                 raise Exception('Error while validation symbol.')

    # def __download_prepared_data(self, data_schema: DataSchema, loop: asyncio.AbstractEventLoop):
    #     for symbol in data_schema.data:
    #         if symbol.additional_properties['files_to_download'] != []:
    #             loop.create_task(self.__download_symbol_data(symbol))

    # def __get_data_source_client(self, historical_source: str) -> DataSource:
    #     client = self.__data_sources[historical_source]
    #     if not client:
    #             self._log('Error, no registered source client')
    #             self.__stop_all_services()
    #     return client

    # def __check_symbol_data_exists(self, data_symbol: DataSymbol) -> List[InstrumentFile]:
    #     """
    #     data scheme
    #     <symbol>__<source>__<interval>__<date-from>__<date-to>
    #     all instruments are downloaded in year files.
    #     """
    #     files: List[InstrumentFile] = get_instrument_files(data_symbol)
    #     # self._log('files to download "'+str(data_symbol.symbol)+'" :', files)
    #     files_in_directory = [f for f in listdir(self.downloaded_data_path) if isfile(join(self.downloaded_data_path, f))]
    #     files_to_download = list(set([f.to_filename() for f in files]) - set(files_in_directory))
    #     files_to_download = [InstrumentFile.from_filename(file) for file in files_to_download]
    #     return files_to_download

    # async def __download_symbol_data(self, symbol: DataSymbol):
    #     files_to_download: List[InstrumentFile] = symbol.additional_properties['files_to_download']
    #     await self.__download_symbol_files(files_to_download, symbol.historical_data_source)

    # async def __download_symbol_files(self, files: List[InstrumentFile], source: str):
    #     data_source_client: DataSource = self.__get_data_source_client(source)
    #     for file in files:
    #         await data_source_client.download_instrument(self.downloaded_data_path, file)

    # def __validate_data_schema_instruments(self, data_symbol_array: List[DataSymbol], loop: asyncio.AbstractEventLoop):
    #     self._log('Data_schema validation')
    #     #check for duplicates:
    #     seen = []
    #     for x in data_symbol_array:
    #         data = str(x.symbol)+str(x.historical_data_source)
    #         if data in seen:
    #             raise Exception('Duplicate data symbols')
    #         else:
    #             seen.append(data)
    #     #check other
    #     number_of_trigger_feeders = 0
    #     for data in data_symbol_array:
    #         if data.backtest_date_start is None:
    #             raise Exception('Error. You must provide "backtest_date_start" field in data_schema file while you are backtesting your strategy')
    #         if data.backtest_date_start >= data.backtest_date_stop:
    #             raise Exception('Error. You have provided "backtest_date_start" is equal or bigger than "backtest_date_start" ')
    #         if [data.backtest_date_start.hour,
    #             data.backtest_date_start.minute,
    #             data.backtest_date_start.second,
    #             data.backtest_date_start.microsecond] != [0,0,0,0]:
    #             raise Exception('Error. Provide your "backtest_date_start" and "backtest_date_stop" in a day accuracy like: "backtest_date_start": datetime(2020,6,1)')
    #         if data.historical_data_source not in self.__historical_sources_array:
    #             raise Exception('Error. This historical_data_source not implemented yet')
    #         if data.trigger_feed == True:
    #             number_of_trigger_feeders += 1
    #     if number_of_trigger_feeders < 1:
    #         raise Exception('Error. Your "data_schema.py" must have at least one instrument that triggers feeds')

    # # override
    # def _loop(self):
    #     loop = asyncio.get_event_loop()
    #     self._create_listeners(loop)
    #     self.__validate_downloaded_data_folder()
    #     if self.config.backtest:
    #         self.__get_data_to_download(self.__data_schema)
    #         self.__create_downloading_clients(self.__historical_sources_array, self.__data_schema, self.__data_sources)
    #         self.__validate_data_schema_instruments(self.__data_schema.data, loop)
    #         self.__validate_symbols_to_download(self.__data_schema, loop)
    #         self.__download_prepared_data(self.__data_schema, loop)

    #         data_parts = self.__prepare_loading_data_structure_2()
    #         loop.create_task(self.__historical_data_loop_ticks(data_parts))
    #     else:
    #         self._log('backtest is off')
    #     loop.run_forever()
    #     loop.close()



#!/usr/bin/env python3.7

from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from typing import Dict, Iterable, List, Optional, TypeVar, Union
from urllib.parse import quote as urlencode

from backtesterrb30.libs.xnt.models.http_jto import (
    Numeric,
    Serializable,
    dc,
    extract_to_model,
    str_to_dt,
    timestamp_to_dt,
)


class InstrumentType(Enum):
    STOCK = "STOCK"
    FUTURE = "FUTURE"
    BOND = "BOND"
    FOREX = "CURRENCY"
    FUND = "FUND"
    OPTION = "OPTION"
    CFD = "CFD"
    CALENDAR_SPREAD = "CALENDAR_SPREAD"
    FX_SPOT = "FX_SPOT"


class Durations(Enum):
    day = "day"
    good_till_cancel = "good_till_cancel"
    good_till_time = "good_till_time"
    immediate_or_cancel = "immediate_or_cancel"
    fill_or_kill = "fill_or_kill"
    at_the_opening = "at_the_opening"
    at_the_close = "at_the_close"
    unknown = "unknown"


class OrderTypes(Enum):
    market = "market"
    limit = "limit"
    stop = "stop"
    stop_limit = "stop_limit"
    twap = "twap"
    iceberg = "iceberg"
    trailing_stop = "trailing_stop"
    unknown = "unknown"


class Side(Enum):
    BUY = "buy"
    SELL = "sell"


class PermissionStatus(Enum):
    blocked = "Blocked"
    full_access = "Full"
    read_only = "ReadOnly"
    close_only = "CloseOnly"

    def __eq__(self, other) -> bool:
        return self.name == other.name

    def __ne__(self, other) -> bool:
        return self.name != other.name


class OptionRight(Enum):
    PULL = "PUT"
    CALL = "CALL"
    BOTH = "BOTH"


class CandleDurations(Enum):
    DAY1 = 86400
    HOUR1 = 3600
    MIN30 = 1800
    MIN5 = 300
    MIN1 = 60


class DataType(Enum):
    QUOTES = "quotes"
    TRADES = "trades"


class FeedLevel(Enum):
    BP = "best_price"
    MD = "market_depth"
    BEST_PRICE = "best_price"
    MARKET_DEPTH = "market_depth"


class Ordering(Enum):
    ASC = "ASC"
    DESC = "DESC"


class OrderStatuses(Enum):
    created = "created"
    accepted = "accepted"
    pending = "pending"
    placing = "placing"
    working = "working"
    cancelled = "cancelled"
    filled = "filled"
    rejected = "rejected"

    @staticmethod
    def terminated(other: Enum):
        return other in (
            OrderStatuses.filled,
            OrderStatuses.rejected,
            OrderStatuses.cancelled,
        )

    @staticmethod
    def active(other: Enum):
        return not OrderStatuses.terminated(other)


class ModifyAction(Enum):
    REPLACE = "replace"
    CANCEL = "cancel"


class Scopes(Enum):
    symbols = "symbols"
    feed = "feed"
    change = "change"
    ohlc = "ohlc"
    crossrates = "crossrates"
    summary = "summary"
    orders = "orders"
    transactions = "transactions"
    accounts = "accounts"


class AuthMethods(Enum):
    JWT = "jwt"
    BASIC = "basic"


class QuoteV1(Serializable):
    def __init__(
        self, symbol_id: str, timestamp: int, bid: Numeric, ask: Numeric
    ) -> None:
        self.timestamp = timestamp_to_dt(timestamp)
        self.symbol_id = symbol_id
        self.bid = dc(bid)
        self.ask = dc(ask)

    @property
    def mid(self) -> Optional[Decimal]:
        if (self.bid is None) or (self.ask is None):
            return None
        else:
            return (self.bid + self.ask) / dc(2)


class QuoteV2(Serializable):
    class Level(Serializable):
        def __init__(self, value: str, size: str) -> None:
            self.value = dc(value)
            self.size = dc(size)

        @property
        def price(self) -> Decimal:
            return self.value

    def __init__(
        self,
        symbol_id: str,
        timestamp: int,
        bid: List[Dict[str, Numeric]],
        ask: List[Dict[str, Numeric]],
    ) -> None:
        self.timestamp = timestamp_to_dt(timestamp)
        self.symbol_id = symbol_id
        self.bid = extract_to_model(bid, self.Level)
        self.ask = extract_to_model(ask, self.Level)

    @property
    def mid(self) -> Optional[Decimal]:
        if (self.bid is None) or (self.ask is None):
            return None
        else:
            return (self.bid[0].price + self.ask[0].price) / dc(2)


class QuoteV3(Serializable):
    class Level(Serializable):
        def __init__(self, price: str, size: str) -> None:
            self.price = dc(price)
            self.size = dc(size)

    def __init__(
        self,
        symbol_id: str,
        timestamp: int,
        bid: List[Dict[str, Numeric]],
        ask: List[Dict[str, Numeric]],
    ) -> None:
        self.timestamp = timestamp_to_dt(timestamp)
        self.symbol_id = symbol_id
        self.bid = extract_to_model(bid, self.Level)
        self.ask = extract_to_model(ask, self.Level)

    @property
    def mid(self) -> Optional[Decimal]:
        if (self.bid is None) or (self.ask is None):
            return None
        else:
            return (self.bid[0].price + self.ask[0].price) / dc(2)


QuoteType = TypeVar("QuoteType", QuoteV1, QuoteV2, QuoteV3, covariant=True)


class TradeV1(Serializable):
    def __init__(self, timestamp: int, symbol_id: str, value: str, size: str) -> None:
        self.timestamp = timestamp_to_dt(timestamp)
        self.symbol_id = symbol_id
        self.value = dc(value)
        self.size = dc(size)


class TradeV2(TradeV1):
    pass


class TradeV3(Serializable):
    def __init__(self, timestamp: int, symbol_id: str, price: str, size: str) -> None:
        self.timestamp = timestamp_to_dt(timestamp)
        self.symbol_id = symbol_id
        self.price = dc(price)
        self.size = dc(size)


TradeType = TypeVar("TradeType", TradeV1, TradeV2, TradeV3, covariant=True)


class ChangeV1(Serializable):
    def __init__(self, base_price: str, daily_change: str, symbol_id: str) -> None:
        self.base_price = dc(base_price)
        self.daily_change = dc(daily_change)
        self.symbol_id = symbol_id


class ChangeV2(ChangeV1):
    pass


class ChangeV3(Serializable):
    def __init__(
        self, last_session_close_price: str, daily_change: str, symbol_id: str
    ) -> None:
        self.last_session_close_price = dc(last_session_close_price)
        self.daily_change = dc(daily_change)
        self.symbol_id = symbol_id


ChangeType = TypeVar("ChangeType", ChangeV1, ChangeV2, ChangeV3, covariant=True)


class OHLCQuotes(Serializable):
    def __init__(
        self,
        open_: Numeric,
        low: Numeric,
        high: Numeric,
        close: Numeric,
        timestamp: int,
    ) -> None:
        self.open_ = dc(open_)
        self.low = dc(low)
        self.high = dc(high)
        self.close = dc(close)
        self.timestamp = timestamp_to_dt(timestamp)


class OHLCTrades(OHLCQuotes):
    def __init__(
        self,
        open_: Numeric,
        low: Numeric,
        high: Numeric,
        close: Numeric,
        timestamp: int,
        volume: Numeric,
    ) -> None:
        super().__init__(open_, low, high, close, timestamp)
        self.volume = dc(volume)


class Crossrate(Serializable):
    def __init__(self, pair: str, symbol_id: str, rate: str) -> None:
        self.pair = pair
        self.symbol_id = symbol_id
        self.rate = rate


class SummaryV1(Serializable):
    class CurrencyPos(Serializable):
        def __init__(self, code: str, value: float, converted_value: float) -> None:
            self.code = code
            self.value = dc(value)
            self.converted_value = dc(converted_value)

    class Position(Serializable):
        def __init__(
            self,
            id_: str,
            symbol_type: str,
            currency: str,
            price: float,
            average_price: float,
            quantity: float,
            value: float,
            converted_value: float,
            pnl: float,
            converted_pnl: float,
        ) -> None:
            self.id_ = id_
            self.symbol_type = Serializable.to_enum(symbol_type, InstrumentType)
            self.currency = currency
            self.price = dc(price)
            self.average_price = dc(average_price)
            self.quantity = dc(quantity)
            self.value = dc(value)
            self.converted_value = dc(converted_value)
            self.pnl = dc(pnl)
            self.converted_pnl = dc(converted_pnl)

    def __init__(
        self,
        account: str,
        timestamp: int,
        currency: str,
        margin_utilization: float,
        free_money: float,
        net_asset_value: float,
        money_used_for_margin: float,
        session_date: str,
        currencies: List[str],
        positions: List[Dict[str, Union[str, float]]],
    ) -> None:
        self.account = account
        self.positions = extract_to_model(positions, self.Position)
        self.currencies = extract_to_model(currencies, self.CurrencyPos)
        self.session_date = session_date
        self.free_money = dc(free_money)
        self.timestamp = timestamp_to_dt(timestamp)
        self.net_asset_value = dc(net_asset_value)
        self.margin_utilization = dc(margin_utilization)
        self.money_used_for_margin = dc(money_used_for_margin)
        self.currency = currency


class SummaryV2(Serializable):
    class CurrencyPos(Serializable):
        def __init__(self, code: str, value: str, converted_value: str) -> None:
            self.code = code
            self.value = dc(value)
            self.converted_value = dc(converted_value)

    class Position(Serializable):
        def __init__(
            self,
            id_: str,
            symbol_type: str,
            currency: str,
            price: str,
            average_price: str,
            quantity: str,
            value: str,
            converted_value: str,
            pnl: str,
            converted_pnl: str,
        ) -> None:
            self.id_ = id_
            self.symbol_type = Serializable.to_enum(symbol_type, InstrumentType)
            self.currency = currency
            self.price = dc(price)
            self.average_price = dc(average_price)
            self.quantity = dc(quantity)
            self.value = dc(value)
            self.converted_value = dc(converted_value)
            self.pnl = dc(pnl)
            self.converted_pnl = dc(converted_pnl)

    def __init__(
        self,
        account: str,
        timestamp: int,
        currency: str,
        margin_utilization: str,
        free_money: str,
        net_asset_value: str,
        money_used_for_margin: str,
        session_date: Optional[List[int]],
        currencies: List[Dict[str, str]],
        positions: List[Dict[str, str]],
    ) -> None:
        self.account = account
        self.positions = extract_to_model(positions, self.Position)
        self.currencies = extract_to_model(currencies, self.CurrencyPos)
        self.session_date = datetime(*session_date) if session_date else None
        self.free_money = dc(free_money)
        self.timestamp = timestamp_to_dt(timestamp)
        self.net_asset_value = dc(net_asset_value)
        self.margin_utilization = dc(margin_utilization)
        self.money_used_for_margin = dc(money_used_for_margin)
        self.currency = currency


class SummaryV3(Serializable):
    class CurrencyPos(Serializable):
        def __init__(self, code: str, price: str, converted_value: str) -> None:
            self.code = code
            self.price = dc(price)
            self.converted_value = dc(converted_value)

    class Position(Serializable):
        def __init__(
            self,
            symbol_id: str,
            symbol_type: str,
            currency: str,
            price: str,
            average_price: str,
            quantity: str,
            value: str,
            converted_value: str,
            pnl: str,
            converted_pnl: str,
        ) -> None:
            self.symbol_id = symbol_id
            self.symbol_type = Serializable.to_enum(symbol_type, InstrumentType)
            self.currency = currency
            self.price = dc(price)
            self.average_price = dc(average_price)
            self.quantity = dc(quantity)
            self.value = dc(value)
            self.converted_value = dc(converted_value)
            self.pnl = dc(pnl)
            self.converted_pnl = dc(converted_pnl)

    def __init__(
        self,
        account_id: str,
        timestamp: int,
        currency: str,
        margin_utilization: float,
        free_money: float,
        net_asset_value: float,
        money_used_for_margin: float,
        session_date: str,
        currencies: List[str],
        positions: List[Dict[str, str]],
    ) -> None:
        self.account_id = account_id
        self.positions = extract_to_model(positions, self.Position)
        self.currencies = extract_to_model(currencies, self.CurrencyPos)
        self.session_date = session_date
        self.free_money = dc(free_money)
        self.timestamp = timestamp_to_dt(timestamp)
        self.net_asset_value = dc(net_asset_value)
        self.margin_utilization = dc(margin_utilization)
        self.money_used_for_margin = dc(money_used_for_margin)
        self.currency = currency


SummaryType = TypeVar("SummaryType", SummaryV1, SummaryV2, SummaryV3, covariant=True)


class TransactionV1(Serializable):
    def __init__(
        self,
        operation_type: str,
        id_: str,
        asset: Optional[str],
        when: int,
        sum_: float,
        symbol_id: Optional[str] = None,
        account_id: Optional[str] = None,
        order_id: Optional[str] = None,
        order_pos: Optional[int] = None,
        uuid_: Optional[str] = None,
        value_date: Optional[str] = None,
    ) -> None:
        self.operation_type = operation_type
        self.id_ = id_
        self.asset = asset
        self.when = timestamp_to_dt(when)
        self.sum_ = sum_
        self.symbol_id = symbol_id
        self.account_id = account_id
        self.order_id = order_id
        self.order_pos = dc(order_pos)
        self.uuid = uuid_
        self.value_date = date.fromisoformat(value_date) if value_date else None  # type: date


class TransactionV2(TransactionV1):
    pass


class TransactionV3(Serializable):
    def __init__(
        self,
        operation_type: Optional[str] = None,
        id_: Optional[str] = None,
        timestamp: Optional[int] = None,
        sum_: Optional[float] = None,
        asset: Optional[str] = None,
        symbol_id: Optional[str] = None,
        account_id: Optional[str] = None,
        order_id: Optional[str] = None,
        order_pos: Optional[int] = None,
        uuid_: Optional[str] = None,
        value_date: Optional[str] = None,
    ) -> None:
        self.operation_type = operation_type
        self.id_ = id_
        self.asset = asset
        self.timestamp = timestamp_to_dt(timestamp)
        self.sum_ = sum_
        self.symbol_id = symbol_id
        self.account_id = account_id
        self.order_id = order_id
        self.order_pos = dc(order_pos)
        self.uuid = uuid_
        self.value_date = date.fromisoformat(value_date) if value_date else None  # type: date


TransactionType = TypeVar(
    "TransactionType", TransactionV1, TransactionV2, TransactionV3, covariant=True
)


class UserAccount(Serializable):
    def __init__(self, status: Union[str, PermissionStatus], account_id: str) -> None:
        self.status = Serializable.to_enum(status, PermissionStatus)
        self.account_id = account_id


class Identifiers(Serializable):
    def __init__(
        self,
        isin: Optional[str] = None,
        figi: Optional[str] = None,
        cusip: Optional[str] = None,
        ric: Optional[str] = None,
        sedol: Optional[str] = None,
    ) -> None:
        self.isin = isin
        self.figi = figi
        self.cusip = cusip
        self.ric = ric
        self.sedol = sedol


class SymbolV1(Serializable):
    class OptionData(Serializable):
        def __init__(
            self, option_group_id: str, right: str, strike_price: float
        ) -> None:
            self.option_group_id = option_group_id
            self.right = Serializable.to_enum(right, OptionRight)
            self.strike_price = dc(strike_price)

    def __init__(
        self,
        name: str,
        description: str,
        country: str,
        exchange: str,
        id_: str,
        currency: str,
        mpi: float,
        type_: str,
        ticker: str,
        group: Optional[str] = None,
        option_data: Optional[Dict[str, str]] = None,
        expiration: Optional[int] = None,
    ) -> None:
        self.option_data = extract_to_model(option_data, self.OptionData)
        self.i18n = {}
        self.name = name
        self.description = description
        self.country = country
        self.exchange = exchange
        self.id_ = id_
        self.currency = currency
        self.mpi = dc(mpi)
        self.type_ = Serializable.to_enum(type_, InstrumentType)
        self.ticker = ticker
        self.expiration = timestamp_to_dt(expiration)
        self.group = group


class SymbolV2(Serializable):
    class OptionData(Serializable):
        def __init__(self, option_group_id: str, right: str, strike_price: str) -> None:
            self.option_group_id = option_group_id
            self.right = Serializable.to_enum(right, OptionRight)
            self.strike_price = dc(strike_price)

    def __init__(
        self,
        name: str,
        description: str,
        country: str,
        exchange: str,
        id_: str,
        currency: str,
        mpi: str,
        type_: str,
        ticker: str,
        group: Optional[str] = None,
        option_data: Optional[Dict[str, str]] = None,
        expiration: Optional[int] = None,
        identifiers: Optional[Dict[str, str]] = None,
    ) -> None:
        self.option_data = extract_to_model(option_data, self.OptionData)
        self.i18n = {}
        self.name = name
        self.description = description
        self.country = country
        self.exchange = exchange
        self.id_ = id_
        self.currency = currency
        self.mpi = dc(mpi)
        self.type_ = Serializable.to_enum(type_, InstrumentType)
        self.ticker = ticker
        self.expiration = timestamp_to_dt(expiration)
        self.group = group
        self.identifiers = extract_to_model(identifiers, Identifiers)


class SymbolV3(Serializable):
    class OptionData(Serializable):
        def __init__(
            self, option_group_id: str, option_right: str, strike_price: str
        ) -> None:
            self.option_group_id = option_group_id
            self.option_right = Serializable.to_enum(option_right, OptionRight)
            self.strike_price = dc(strike_price)

    def __init__(
        self,
        name: str,
        description: str,
        country: str,
        exchange: str,
        symbol_id: str,
        currency: str,
        min_price_increment: str,
        symbol_type: str,
        ticker: str,
        group: str,
        expiration: Optional[int] = None,
        option_data: Optional[Dict[str, str]] = None,
        underlying_symbol_id: Optional[str] = None,
        identifiers: Optional[Dict[str, str]] = None,
        icon: Optional[str] = None,
    ) -> None:
        self.option_data = extract_to_model(option_data, self.OptionData)
        self.name = name
        self.description = description
        self.country = country
        self.exchange = exchange
        self.symbol_id = symbol_id
        self.currency = currency
        self.min_price_increment = dc(min_price_increment)
        self.symbol_type = Serializable.to_enum(symbol_type, InstrumentType)
        self.ticker = ticker
        self.expiration = timestamp_to_dt(expiration)
        self.group = group
        self.underlying_symbol_id = underlying_symbol_id
        self.identifiers = extract_to_model(identifiers, Identifiers)
        self.icon = icon


SymbolType = TypeVar("SymbolType", SymbolV1, SymbolV2, SymbolV3, covariant=True)


class SymbolSpecification(Serializable):
    def __init__(
        self,
        leverage: Numeric,
        contract_multiplier: Numeric,
        price_unit: Numeric,
        units: str,
        lot_size: Numeric,
    ) -> None:
        self.leverage = dc(leverage)
        self.contract_multiplier = dc(contract_multiplier)
        self.price_unit = dc(price_unit)
        self.units = units
        self.lot_size = dc(lot_size)


class Exchange(Serializable):
    def __init__(self, id_: str, name: str, country: str) -> None:
        self.id_ = id_
        self.name = name
        self.country = country


class Group(Serializable):
    def __init__(self, group: str, name: str, types: List[str], exchange: str) -> None:
        self.group = group
        self.name = name
        self.types = [Serializable.to_enum(ty, InstrumentType) for ty in types]
        self.exchange = exchange


class Schedule(Serializable):
    class Interval(Serializable):
        class Period(Serializable):
            def __init__(self, start: int, end: int) -> None:
                self.start = timestamp_to_dt(start)
                self.end = timestamp_to_dt(end)

        def __init__(
            self,
            name: str,
            period: Dict[str, int],
            order_types: Optional[Dict[str, str]] = None,
        ) -> None:
            self.name = name
            self.period = extract_to_model(period, self.Period)
            self.order_types = order_types

        @property
        def start(self) -> datetime:
            return self.period.start

        @property
        def end(self) -> datetime:
            return self.period.end

    def __init__(self, intervals: list) -> None:
        self.intervals = extract_to_model(intervals, self.Interval)


class OrderSentV1(Serializable):
    order_type = None  # type: OrderTypes

    def __init__(
        self,
        account: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
    ) -> None:
        self.account = account
        self.instrument = instrument
        self.side = side
        self.quantity = dc(quantity)
        self.duration = duration
        self.client_tag = client_tag
        self.oco_group = oco_group
        self.if_done_parent_id = if_done_parent_id


class OrderSentV2(Serializable):
    order_type = None  # type: OrderTypes

    def __init__(
        self,
        account_id: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
        take_profit: Optional[Numeric] = None,
        stop_loss: Optional[Numeric] = None,
    ) -> None:
        self.account_id = account_id
        self.instrument = instrument
        self.side = side
        self.quantity = dc(quantity)
        self.duration = duration
        self.client_tag = client_tag
        self.oco_group = oco_group
        self.if_done_parent_id = if_done_parent_id
        self.take_profit = take_profit
        self.stop_loss = stop_loss


class OrderSentV3(Serializable):
    order_type = None  # type: OrderTypes

    def __init__(
        self,
        account_id: str,
        symbol_id: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
        take_profit: Optional[Numeric] = None,
        stop_loss: Optional[Numeric] = None,
    ) -> None:
        self.account_id = account_id
        self.symbol_id = symbol_id
        self.side = side
        self.quantity = dc(quantity)
        self.duration = duration
        self.client_tag = client_tag
        self.oco_group = oco_group
        self.if_done_parent_id = if_done_parent_id
        self.take_profit = take_profit
        self.stop_loss = stop_loss


OrderSentType = TypeVar(
    "OrderSentType", OrderSentV1, OrderSentV2, OrderSentV3, covariant=True
)


class OrderMarketV1(OrderSentV1):
    def __init__(
        self,
        account: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
    ) -> None:
        super().__init__(
            account,
            instrument,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
        )
        self.order_type = OrderTypes.market


class OrderMarketV2(OrderSentV2):
    def __init__(
        self,
        account_id: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
        take_profit: Optional[Numeric] = None,
        stop_loss: Optional[Numeric] = None,
    ) -> None:
        super().__init__(
            account_id,
            instrument,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
            take_profit,
            stop_loss,
        )
        self.order_type = OrderTypes.market


class OrderMarketV3(OrderSentV3):
    def __init__(
        self,
        account_id: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
        take_profit: Optional[Numeric] = None,
        stop_loss: Optional[Numeric] = None,
    ) -> None:
        super().__init__(
            account_id,
            instrument,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
            take_profit,
            stop_loss,
        )
        self.order_type = OrderTypes.market


OrderMarketType = TypeVar(
    "OrderMarketType", OrderMarketV1, OrderMarketV2, OrderMarketV3, covariant=True
)


class OrderLimitV1(OrderSentV1):
    def __init__(
        self,
        account: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        limit_price: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
    ) -> None:
        super().__init__(
            account,
            instrument,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
        )
        self.order_type = OrderTypes.limit
        self.limit_price = dc(limit_price)


class OrderLimitV2(OrderSentV2):
    def __init__(
        self,
        account_id: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        limit_price: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
        take_profit: Optional[Numeric] = None,
        stop_loss: Optional[Numeric] = None,
    ) -> None:
        super().__init__(
            account_id,
            instrument,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
            take_profit,
            stop_loss,
        )
        self.order_type = OrderTypes.limit
        self.limit_price = dc(limit_price)


class OrderLimitV3(OrderSentV3):
    def __init__(
        self,
        account_id: str,
        symbol_id: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        limit_price: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
        take_profit: Optional[Numeric] = None,
        stop_loss: Optional[Numeric] = None,
    ) -> None:
        super().__init__(
            account_id,
            symbol_id,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
            take_profit,
            stop_loss,
        )
        self.order_type = OrderTypes.limit
        self.limit_price = limit_price


OrderLimitType = TypeVar(
    "OrderLimitType", OrderLimitV1, OrderLimitV2, OrderLimitV3, covariant=True
)


class OrderStopV1(OrderSentV1):
    def __init__(
        self,
        account: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        stop_price: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
    ) -> None:
        super().__init__(
            account,
            instrument,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
        )
        self.order_type = OrderTypes.stop
        self.stop_price = stop_price


class OrderStopV2(OrderSentV2):
    def __init__(
        self,
        account_id: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        stop_price: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
        take_profit: Optional[Numeric] = None,
        stop_loss: Optional[Numeric] = None,
    ) -> None:
        super().__init__(
            account_id,
            instrument,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
            take_profit,
            stop_loss,
        )
        self.order_type = OrderTypes.stop
        self.stop_price = stop_price


class OrderStopV3(OrderSentV3):
    def __init__(
        self,
        account_id: str,
        symbol_id: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        stop_price: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
        take_profit: Optional[Numeric] = None,
        stop_loss: Optional[Numeric] = None,
    ) -> None:
        super().__init__(
            account_id,
            symbol_id,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
            take_profit,
            stop_loss,
        )
        self.order_type = OrderTypes.stop
        self.stop_price = stop_price


OrderStopType = TypeVar(
    "OrderStopType", OrderStopV1, OrderStopV2, OrderStopV3, covariant=True
)


class OrderStopLimitV1(OrderSentV1):
    def __init__(
        self,
        account: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        limit_price: Numeric,
        stop_price: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
    ) -> None:
        super().__init__(
            account,
            instrument,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
        )
        self.order_type = OrderTypes.stop_limit
        self.limit_price = limit_price
        self.stop_price = stop_price


class OrderStopLimitV2(OrderSentV2):
    def __init__(
        self,
        account_id: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        limit_price: Numeric,
        stop_price: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
        take_profit: Optional[Numeric] = None,
        stop_loss: Optional[Numeric] = None,
    ) -> None:
        super().__init__(
            account_id,
            instrument,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
            take_profit,
            stop_loss,
        )
        self.order_type = OrderTypes.stop_limit
        self.limit_price = limit_price
        self.stop_price = stop_price


class OrderStopLimitV3(OrderSentV3):
    def __init__(
        self,
        account_id: str,
        symbol_id: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        limit_price: Numeric,
        stop_price: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
        take_profit: Optional[Numeric] = None,
        stop_loss: Optional[Numeric] = None,
    ) -> None:
        super().__init__(
            account_id,
            symbol_id,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
            take_profit,
            stop_loss,
        )
        self.order_type = OrderTypes.stop_limit
        self.limit_price = limit_price
        self.stop_price = stop_price


OrderStopLimitType = TypeVar(
    "OrderStopLimitType",
    OrderStopLimitV1,
    OrderStopLimitV2,
    OrderStopLimitV3,
    covariant=True,
)


class OrderTrailingStopV1(OrderSentV1):
    def __init__(
        self,
        account: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        price_distance: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
    ) -> None:
        super().__init__(
            account,
            instrument,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
        )
        self.order_type = OrderTypes.trailing_stop
        self.price_distance = price_distance


class OrderTrailingStopV2(OrderSentV2):
    def __init__(
        self,
        account_id: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        price_distance: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
        take_profit: Optional[Numeric] = None,
        stop_loss: Optional[Numeric] = None,
    ) -> None:
        super().__init__(
            account_id,
            instrument,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
            take_profit,
            stop_loss,
        )
        self.order_type = OrderTypes.trailing_stop
        self.price_distance = price_distance


class OrderTrailingStopV3(OrderSentV3):
    def __init__(
        self,
        account_id: str,
        symbol_id: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        price_distance: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
        take_profit: Optional[Numeric] = None,
        stop_loss: Optional[Numeric] = None,
    ) -> None:
        super().__init__(
            account_id,
            symbol_id,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
            take_profit,
            stop_loss,
        )
        self.order_type = OrderTypes.trailing_stop
        self.price_distance = price_distance


OrderTrailingStopType = TypeVar(
    "OrderTrailingStopType",
    OrderTrailingStopV1,
    OrderTrailingStopV2,
    OrderTrailingStopV3,
    covariant=True,
)


class OrderTwapV1(OrderSentV1):
    def __init__(
        self,
        account: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        part_quantity: Numeric,
        place_interval: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
    ) -> None:
        super().__init__(
            account,
            instrument,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
        )
        self.order_type = OrderTypes.twap
        self.part_quantity = part_quantity
        self.place_interval = place_interval


class OrderTwapV2(OrderSentV2):
    def __init__(
        self,
        account_id: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        part_quantity: Numeric,
        place_interval: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
        take_profit: Optional[Numeric] = None,
        stop_loss: Optional[Numeric] = None,
    ) -> None:
        super().__init__(
            account_id,
            instrument,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
            take_profit,
            stop_loss,
        )
        self.order_type = OrderTypes.twap
        self.part_quantity = part_quantity
        self.place_interval = place_interval


class OrderTwapV3(OrderSentV3):
    def __init__(
        self,
        account_id: str,
        symbol_id: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        part_quantity: Numeric,
        place_interval: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
        take_profit: Optional[Numeric] = None,
        stop_loss: Optional[Numeric] = None,
    ) -> None:
        super().__init__(
            account_id,
            symbol_id,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
            take_profit,
            stop_loss,
        )
        self.order_type = OrderTypes.twap
        self.part_quantity = part_quantity
        self.place_interval = place_interval


OrderTwapType = TypeVar(
    "OrderTwapType", OrderTwapV1, OrderTwapV2, OrderTwapV3, covariant=True
)


class OrderIcebergV1(OrderSentV1):
    def __init__(
        self,
        account: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        part_quantity: Numeric,
        limit_price: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
    ) -> None:
        super().__init__(
            account,
            instrument,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
        )
        self.order_type = OrderTypes.iceberg
        self.limit_price = limit_price
        self.part_quantity = part_quantity


class OrderIcebergV2(OrderSentV2):
    def __init__(
        self,
        account_id: str,
        instrument: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        part_quantity: Numeric,
        limit_price: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
        take_profit: Optional[Numeric] = None,
        stop_loss: Optional[Numeric] = None,
    ) -> None:
        super().__init__(
            account_id,
            instrument,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
            take_profit,
            stop_loss,
        )
        self.order_type = OrderTypes.iceberg
        self.limit_price = limit_price
        self.part_quantity = part_quantity


class OrderIcebergV3(OrderSentV3):
    def __init__(
        self,
        account_id: str,
        symbol_id: str,
        side: Side,
        quantity: Numeric,
        duration: Durations,
        part_quantity: Numeric,
        limit_price: Numeric,
        client_tag: Optional[str] = None,
        oco_group: Optional[str] = None,
        if_done_parent_id: Optional[str] = None,
        take_profit: Optional[Numeric] = None,
        stop_loss: Optional[Numeric] = None,
    ) -> None:
        super().__init__(
            account_id,
            symbol_id,
            side,
            quantity,
            duration,
            client_tag,
            oco_group,
            if_done_parent_id,
            take_profit,
            stop_loss,
        )
        self.order_type = OrderTypes.iceberg
        self.limit_price = limit_price
        self.part_quantity = part_quantity


OrderIcebergType = TypeVar(
    "OrderIcebergType", OrderIcebergV1, OrderIcebergV2, OrderIcebergV3, covariant=True
)


class Fill(Serializable):
    def __init__(self, quantity: str, price: str, position: int, time: str) -> None:
        self.quantity = dc(quantity)
        self.price = dc(price)
        self.position = position
        self.time = str_to_dt(time)


class OrderState(Serializable):
    def __init__(
        self,
        last_update: str,
        status: str,
        fills: List[Dict[str, Union[str, int]]],
        reason: Optional[str] = None,
    ):
        self.last_update = str_to_dt(last_update)
        self.status = Serializable.to_enum(status, OrderStatuses)
        self.fills = extract_to_model(fills, Fill)
        # TODO: rewrite it to normal Reject model
        self.reason = reason


class Reject(Serializable):
    def __init__(self, group: str, message: str):
        self.group = group
        self.message = message


class OrderV1(Serializable):
    class OrderParameters(Serializable):
        def __init__(
            self,
            side: str,
            duration: str,
            quantity: str,
            instrument: str,
            order_type: str,
            oco_group: Optional[str] = None,
            if_done_parent_id: Optional[str] = None,
            limit_price: Optional[str] = None,
            stop_price: Optional[str] = None,
            price_distance: Optional[str] = None,
            part_quantity: Optional[str] = None,
            place_interval: Optional[str] = None,
        ) -> None:
            self.side = Serializable.to_enum(side, Side)
            self.duration = Serializable.to_enum(duration, Durations)
            self.quantity = dc(quantity)
            self.instrument = instrument
            self.order_type = Serializable.to_enum(order_type, OrderTypes)
            self.oco_group = oco_group
            self.if_done_parent_id = if_done_parent_id
            self.limit_price = dc(limit_price)
            self.stop_price = dc(stop_price)
            self.price_distance = dc(price_distance)
            self.part_quantity = dc(part_quantity)
            self.place_interval = dc(place_interval)

    def __init__(
        self,
        place_time: str,
        order_state: Dict[str, Union[List, str, None]],
        id_: str,
        order_parameters: Dict[str, Optional[str]],
        current_modification_id: str,
        exante_account: str,
        client_tag: Optional[str] = None,
        username: Optional[str] = None,
    ):
        self.place_time = str_to_dt(place_time)
        self.order_state = extract_to_model(order_state, OrderState)
        self.id_ = id_
        self.client_tag = client_tag
        self.order_parameters = extract_to_model(order_parameters, self.OrderParameters)
        self.current_modification_id = current_modification_id
        self.exante_account = exante_account
        self.username = username


class OrderV2(Serializable):
    class OrderParameters(Serializable):
        def __init__(
            self,
            side: str,
            duration: str,
            quantity: str,
            instrument: str,
            order_type: str,
            oco_group: Optional[str] = None,
            if_done_parent_id: Optional[str] = None,
            limit_price: Optional[str] = None,
            stop_price: Optional[str] = None,
            price_distance: Optional[str] = None,
            part_quantity: Optional[str] = None,
            place_interval: Optional[str] = None,
        ) -> None:
            self.side = Serializable.to_enum(side, Side)
            self.duration = Serializable.to_enum(duration, Durations)
            self.quantity = dc(quantity)
            self.instrument = instrument
            self.order_type = Serializable.to_enum(order_type, OrderTypes)
            self.oco_group = oco_group
            self.if_done_parent_id = if_done_parent_id
            self.limit_price = dc(limit_price)
            self.stop_price = dc(stop_price)
            self.price_distance = dc(price_distance)
            self.part_quantity = dc(part_quantity)
            self.place_interval = dc(place_interval)

    def __init__(
        self,
        place_time: str,
        order_state: Dict,
        id_: str,
        order_parameters: Dict,
        username: str,
        current_modification_id: str,
        account_id: str = None,
        client_tag: Optional[str] = None,
    ) -> None:
        self.place_time = str_to_dt(place_time)
        self.order_state = extract_to_model(order_state, OrderState)
        self.id_ = id_
        self.client_tag = client_tag
        self.order_parameters = extract_to_model(order_parameters, self.OrderParameters)
        self.current_modification_id = current_modification_id
        self.account_id = account_id
        self.username = username


class OrderV3(Serializable):
    class OrderParameters(Serializable):
        def __init__(
            self,
            side: str,
            duration: str,
            quantity: str,
            symbol_id: str,
            order_type: str,
            oco_group: Optional[str] = None,
            if_done_parent_id: Optional[str] = None,
            limit_price: Optional[str] = None,
            stop_price: Optional[str] = None,
            price_distance: Optional[str] = None,
            part_quantity: Optional[str] = None,
            place_interval: Optional[str] = None,
        ) -> None:
            self.side = Serializable.to_enum(side, Side)
            self.duration = Serializable.to_enum(duration, Durations)
            self.quantity = dc(quantity)
            self.symbol_id = symbol_id
            self.order_type = Serializable.to_enum(order_type, OrderTypes)
            self.oco_group = oco_group
            self.if_done_parent_id = if_done_parent_id
            self.limit_price = dc(limit_price)
            self.stop_price = dc(stop_price)
            self.price_distance = dc(price_distance)
            self.part_quantity = dc(part_quantity)
            self.place_interval = dc(place_interval)

    def __init__(
        self,
        place_time: str,
        order_state: Dict,
        order_id: str,
        order_parameters: Dict,
        username: str,
        current_modification_id: str,
        account_id: str,
        client_tag: Optional[str] = None,
    ) -> None:
        self.place_time = str_to_dt(place_time)
        self.order_state = extract_to_model(order_state, OrderState)
        self.order_id = order_id
        self.client_tag = client_tag
        self.order_parameters = extract_to_model(order_parameters, self.OrderParameters)
        self.current_modification_id = current_modification_id
        self.account_id = account_id
        self.username = username


OrderType = TypeVar("OrderType", OrderV1, OrderV2, OrderV3, covariant=True)


class ExOrderV1(Serializable):
    def __init__(
        self,
        quantity: str,
        order_id: str,
        event: str,
        price: str,
        position: str,
        time: str,
    ) -> None:
        self.quantity = dc(quantity)
        self.order_id = order_id
        self.event = event
        self.price = dc(price)
        self.position = dc(position)
        self.time = str_to_dt(time)


class ExOrderV2(ExOrderV1):
    pass


class ExOrderV3(Serializable):
    def __init__(
        self, quantity: str, order_id: str, price: str, position: str, timestamp: str
    ) -> None:
        self.quantity = dc(quantity)
        self.order_id = order_id
        self.price = dc(price)
        self.position = dc(position)
        self.timestamp = str_to_dt(timestamp)


ExOrderType = TypeVar("ExOrderType", ExOrderV1, ExOrderV2, ExOrderV3, covariant=True)


def resolve_model(
    version: str,
    class_type: Union[
        TradeType,
        QuoteType,
        ChangeType,
        TradeType,
        SummaryType,
        OrderSentType,
        OrderMarketType,
        OrderLimitType,
        OrderStopType,
        OrderStopLimitType,
        OrderTrailingStopType,
        OrderTwapType,
        OrderIcebergType,
    ],
):
    classes = list(
        filter(
            lambda x: f"{class_type.__name__[:-4]}V{version.split('.')[0]}"
            == x.__name__,
            class_type.__constraints__,
        )
    )  # type: List[Serializable]
    if len(classes) > 1:
        raise Exception("Unexpected resolution!")
    else:
        return classes[0]


def resolve_symbol(
    symbol: Union[str, SymbolType, Iterable[str], Iterable[SymbolType]],
) -> Optional[str]:
    if isinstance(symbol, SymbolV3):
        return urlencode(symbol.symbol_id, safe="")
    elif isinstance(symbol, (SymbolV1, SymbolV2)):
        return urlencode(symbol.id_, safe="")
    elif isinstance(symbol, Iterable) and not isinstance(symbol, str):
        return ",".join([resolve_symbol(x) for x in symbol])
    elif symbol:
        return urlencode(symbol, safe="")
    else:
        return symbol



import asyncio
import time

from appdirs import user_cache_dir

from backtesterrb30.historical_data_feeds.functions import load_data_frame_with_dfs
from backtesterrb30.libs.communication_broker.broker_base import BrokerBase
from backtesterrb30.libs.data_sources.data_source_base import DataSource
from backtesterrb30.libs.data_sources.data_sources_list import HISTORICAL_SOURCES
from backtesterrb30.libs.interfaces.utils.config import Config
from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.utils.list_of_services import SERVICES
from backtesterrb30.libs.utils.service import Service


class LiveDataFeeds(Service):
    """Python Live data provider"""

    downloaded_data_path = user_cache_dir("rb30_cache")
    _broker: BrokerBase

    def __init__(self, config: dict, data_schema: DataSchema, loop=None, logger=print):
        super().__init__(config, logger)
        self.config: Config = config
        self.__loop = loop
        self.__custom_event_loop = False
        if self.__loop is None:
            self.__loop = asyncio.get_event_loop()
            self.__custom_event_loop = True
        self.__data_schema: DataSchema = data_schema
        self.__columns = ["timestamp"] + [c.symbol for c in self.__data_schema.data]
        self.__historical_sources_array = [
            i for i in dir(HISTORICAL_SOURCES) if not i.startswith("__")
        ]
        self.__data_sources = {}
        self.__download_buffer = True
        self.__engine_ready = False
        self.__historical_rows = []

    # override
    def _loop(self):
        self._broker.create_listeners(self.__loop)
        # TODO add data downloader
        if not self.config.backtest:
            self.__create_downloading_clients(
                self.__historical_sources_array,
                self.__data_schema,
                self.__data_sources,
                self._log,
            )
            self.__combine_clients_and_symbols(self.__data_schema)
            interval = self.__find_loop_interval(self.__data_schema)
            self.__loop.create_task(self.__main_loop(interval))
        if self.__custom_event_loop:
            self.__loop.run_forever()
            self.__loop.close()

    def _configure(self):
        super()._configure()
        self._broker.register(
            "engine_set_buffer_length", self.__engine_set_buffer_length_event
        )

    # def _send(self, service: SERVICES, msg: str, *args):
    #     self._broker.send(service, msg, *args)

    # # override
    # def _handle_zmq_message(self, message):
    #     pass

    async def __main_loop(self, interval):
        if self.__download_buffer:
            await self._broker.send(
                SERVICES.python_engine,
                "get_buffer_length",
                SERVICES.live_data_feeds.value,
            )
        else:
            self.__engine_ready = True
        while True:
            if self.__engine_ready:
                timestamp = await self.__get_proper_timestamp(
                    self.__historical_rows, self.__data_schema
                )
                for row in self.__historical_rows:
                    last_row = row
                    await self._broker.send(
                        SERVICES.python_engine, "data_feed", list(last_row)
                    )
                while True:
                    row = []
                    row.append(timestamp)
                    sth_updated = False
                    for data_symbol in self.__data_schema.data:
                        price, updated = self.__get_symbol_price(timestamp, data_symbol)
                        if updated:
                            sth_updated = True
                        row.append(price)
                    if sth_updated:
                        await self._broker.send(
                            SERVICES.python_engine, "data_feed", row
                        )
                    await asyncio.sleep(interval / 1000)
                    timestamp += interval
            await asyncio.sleep(2)

    def __get_smallest_interval(self, data_schema: DataSchema):
        smallest = 9999999999999
        for data_symbol in data_schema.data:
            client: DataSource = data_symbol.additional_properties["downloading_client"]
            milis = client._get_interval_miliseconds(data_symbol.interval.value)
            if not milis or not (1000 * 60 <= milis <= 1000 * 60 * 60 * 24 * 6):
                raise Exception(data_symbol.symbol, "Cant run live with this interval")
            if milis < smallest:
                smallest = milis
        return smallest

    async def __download_data(self, buffer_length: int, data_schema: DataSchema):
        buffer_length = buffer_length
        smallest = self.__get_smallest_interval(data_schema)
        time_stop = int(time.time()) * 1000 + smallest
        time_start = time_stop - 4 * smallest * buffer_length
        if time_stop - time_start < 1000 * 60 * 60 * 24:
            time_start = time_stop - 1000 * 60 * 60 * 24
        dfs = []
        for data_symbol in data_schema.data:
            client: DataSource = data_symbol.additional_properties["downloading_client"]
            df = await client.download_dataframe(
                data_symbol.symbol, data_symbol.interval.value, time_start, None
            )
            df.columns = ["timestamp", data_symbol.identifier]
            dfs.append(df)
        self.__historical_rows = load_data_frame_with_dfs(
            self.__data_schema, self.__columns, [], dfs
        )
        self.__historical_rows = self.__historical_rows[-buffer_length - 1 :]

    async def __get_proper_timestamp(self, historical_rows, data_schema: DataSchema):
        if historical_rows == []:
            return time.time() * 1000
        smallest = self.__get_smallest_interval(data_schema)
        current_time_with_reserve = time.time() * 1000 + 1000
        next_should_be_timestamp = historical_rows[-1][0] + smallest
        if next_should_be_timestamp < current_time_with_reserve:
            raise Exception(
                "Did not managed to download historical buffer this time. Trying again"
            )
        await asyncio.sleep(abs(next_should_be_timestamp - time.time() * 1000) / 1000)
        return next_should_be_timestamp

    def __find_loop_interval(self, data_schema: DataSchema) -> int:
        smallest = self.__get_smallest_interval(data_schema)
        for data_symbol in data_schema.data:
            client: DataSource = data_symbol.additional_properties["downloading_client"]
            milis = client._get_interval_miliseconds(data_symbol.interval.value)
            num = (milis / smallest) % 1000
            if int(num) != num:
                raise Exception(
                    "Bad intervals... All intervals should be divisible by smallest one."
                )
        return smallest

    def __get_symbol_price(self, timestamp: int, symbol: DataSymbol):
        updated = False
        price = 0
        client: DataSource = symbol.additional_properties["downloading_client"]
        if "last_price_timestamp" not in symbol.additional_properties:
            symbol.additional_properties["last_price_timestamp"] = timestamp
            updated = True
        if "last_price" not in symbol.additional_properties:
            symbol.additional_properties["last_price"] = client.get_current_price(
                symbol
            )
            updated = True
        milis = client._get_interval_miliseconds(symbol.interval.value)
        if timestamp - symbol.additional_properties["last_price_timestamp"] < milis:
            price = symbol.additional_properties["last_price"]
        if timestamp - symbol.additional_properties["last_price_timestamp"] == milis:
            price = client.get_current_price(symbol)
            symbol.additional_properties["last_price_timestamp"] = timestamp
            symbol.additional_properties["last_price"] = price
            updated = True
        if timestamp - symbol.additional_properties["last_price_timestamp"] > milis:
            raise Exception("Error. Bad timestamp Such situation should not happen")
        return price, updated

    def __create_downloading_clients(
        self,
        historical_sources_array: list,
        data_schema: DataSchema,
        data_sources: dict,
        log=print,
    ):
        for source in historical_sources_array:
            if source in [data.historical_data_source for data in data_schema.data]:
                data_sources[source]: DataSource = getattr(HISTORICAL_SOURCES, source)(
                    log
                )

    def __combine_clients_and_symbols(self, data_schema: DataSchema):
        for data_symbol in data_schema.data:
            data_symbol.additional_properties["downloading_client"] = (
                self.__get_data_source_client(data_symbol.historical_data_source)
            )

    def __get_data_source_client(self, historical_source: str) -> DataSource:
        client = self.__data_sources[historical_source]
        if not client:
            raise Exception("Error, no registered source client")
        return client

    # COMMANDS

    async def __engine_set_buffer_length_event(self, buffer_length: int):
        self.__buffer_length = buffer_length
        await self.__download_data(buffer_length, self.__data_schema)
        self.__engine_ready = True



#!/usr/bin/env python3.7

import json
import logging
from datetime import datetime, timezone
from queue import Empty, Queue
from threading import Event, Thread
from time import sleep
from typing import Any, Callable, Dict, Iterable, List, Optional, Type, Union
from urllib.parse import quote as urlencode

import backoff
import jwt
import requests
from requests import adapters, exceptions
from requests.auth import AuthBase, HTTPBasicAuth

from backtesterrb30.libs.xnt.models.http_api_models import (
    AuthMethods,
    CandleDurations,
    ChangeType,
    Crossrate,
    DataType,
    Exchange,
    ExOrderType,
    FeedLevel,
    Group,
    InstrumentType,
    ModifyAction,
    OHLCQuotes,
    OHLCTrades,
    Ordering,
    OrderSentType,
    OrderType,
    OrderV1,
    OrderV2,
    OrderV3,
    QuoteType,
    Reject,
    Schedule,
    Scopes,
    SummaryType,
    SymbolSpecification,
    SymbolType,
    TradeType,
    TransactionType,
    UserAccount,
    resolve_model,
    resolve_symbol,
)
from backtesterrb30.libs.xnt.models.http_jto import (
    Numeric,
    SerializableType,
    attr_or,
    dt_to_str,
    dt_to_timestamp,
    extract_to_model,
    opt_int,
    timestamp_to_dt,
)

versions = ("1.0", "2.0", "3.0")
current_api = "2.0"


class JWTAuth(Thread, AuthBase):
    def __init__(
        self,
        appid: str,
        client_id: str,
        shared_key: str,
        ttl: Numeric,
        scopes: Iterable[Scopes],
    ):
        self.appid = appid
        self.client_id = client_id
        self.shared_key = shared_key
        self.ttl = int(ttl)
        self.scopes = scopes
        self.token = None
        super().__init__(daemon=True)
        self.start()

    def make_token(self) -> str:
        now = int(datetime.now(tz=timezone.utc).timestamp())
        token = jwt.encode(
            payload={
                "iss": self.client_id,
                "sub": self.appid,
                "iat": now,
                "exp": now + self.ttl,
                "aud": [x.value for x in self.scopes],
            },
            key=self.shared_key,
            algorithm="HS256",
        ).decode("ascii")
        return token

    def run(self):
        while True:
            self.token = self.make_token()
            sleep(self.ttl)

    def __repr__(self):
        return f"Bearer {self.token}"

    def __call__(self, r):
        r.headers["Authorization"] = f"Bearer {self.token}"
        return r


def conerror(exc):
    exception = []
    print(exc, type(exc))
    return any([isinstance(exc, x) for x in exception])


class HTTPApiStreaming(Thread):
    def __init__(
        self,
        auth: AuthBase,
        api: str,
        handler: str,
        model: Type[SerializableType],
        params: Optional[Dict[str, Any]] = None,
        event_filter: Optional[str] = None,
        chunk_size: Optional[int] = 1,
        session: Optional[requests.Session] = None,
        logger: Optional[logging.Logger] = None,
        log_level: Optional[int] = logging.ERROR,
        backup_model: Optional[Type[SerializableType]] = None,
        debug_raise_model_exceptions: bool = False,
    ) -> None:
        self.auth = auth
        self.api = api
        self.handler = handler
        self.chunk_size = chunk_size
        self.params = params
        self.model = model
        self.backup_model = backup_model
        self.session = session or requests.Session()
        self.session.mount(api, adapters.HTTPAdapter())
        self.is_finished = Event()
        self.queue = Queue(maxsize=0)
        self.event_filter = event_filter
        self._eraise = debug_raise_model_exceptions
        super().__init__(daemon=True)
        self.logger = logger or logging.Logger(name=self.name, level=log_level)
        self.start()

    def run(self) -> None:
        headers = {"Accept": "application/x-json-stream", "Accept-Encoding": "gzip"}
        mes: Dict = {}
        while not self.is_finished.is_set():
            for message in self.session.get(
                self.api + self.handler,
                headers=headers,
                stream=True,
                params=self.params,
                timeout=60,
                auth=self.auth,
            ).iter_lines(chunk_size=self.chunk_size):
                if self.is_finished.is_set():
                    break
                try:
                    mes = json.loads(message.decode())
                    if mes.get("event") == "heartbeat":
                        continue
                    elif self.event_filter and self.event_filter == mes.get("event"):
                        parsed_message = extract_to_model(
                            mes[self.event_filter],
                            self.model,
                            self._eraise,
                            backup_obj=self.backup_model,
                        )
                    else:
                        parsed_message = extract_to_model(
                            mes, self.model, self._eraise, backup_obj=self.backup_model
                        )
                    if parsed_message is not None:
                        self.queue.put(parsed_message)
                except json.JSONDecodeError as exc:
                    if self._eraise:
                        raise exc
                    else:
                        self.logger.error(
                            "Unable to parse message as JSON: %s" % message
                        )
                        continue
                except RuntimeError as exc:
                    if self._eraise:
                        raise exc
                    else:
                        self.logger.error(
                            "%s found at %s: '%s'" % (exc.__class__, mes, exc.args[0])
                        )
                        continue

    def get(
        self, block: bool = False, timeout: int = 30, eraise: bool = False
    ) -> Optional[SerializableType]:
        try:
            return self.queue.get(block=block, timeout=timeout)
        except Empty:
            if not eraise:
                return None
            else:
                raise Empty

    def stop(self) -> None:
        self.is_finished.set()
        self.join()


class HTTPApi:
    def __init__(
        self,
        auth: AuthMethods,
        appid: str,
        acckey: Optional[str] = None,
        clientid: Optional[str] = None,
        sharedkey: Optional[str] = None,
        ttl: Optional[int] = None,
        scopes: Iterable[Scopes] = tuple(Scopes),
        url: str = "https://api-live.exante.eu",
        version: str = current_api,
        logger: Optional[logging.Logger] = None,
        log_level: Optional[int] = logging.ERROR,
        debug_raise_model_exceptions: bool = False,
    ) -> None:
        """
        :param auth: auth - basic or jwt. Auth method
        :param appid: ApplicationId
        :param acckey: [Basic] AccessKey
        :param clientid: [JWT] ClientId
        :param sharedkey: [JWT] Shared Key for signing
        :param ttl: [JWT] token time to live in seconds, default is 1 hour
        :param scopes: [JWT] scopes token accessible to, reffer to specs, default is full
        :param url: url to connect to
        :param version: default API version, used if not overriden in call
        :param logger: external logging.Logger object
        :param log_level: log_level int code (50 for ERROR) or logging.LOG_LEVEL const
        :param debug_raise_model_exceptions: raise RuntimeError on failures on model serializing
        (default will return None and warning)
        Usefull on API changes, testing or forking
        """
        if version in versions:
            self.version = version
        else:
            raise ValueError(f"Version should be one of {versions}")
        self._eraise = debug_raise_model_exceptions
        if auth == AuthMethods.BASIC:
            if appid and acckey:
                self.auth = HTTPBasicAuth(appid, acckey)
            else:
                raise ValueError(
                    "applicationId and accountKey must be provided for basic authentification"
                )
        elif auth == AuthMethods.JWT:
            if appid and clientid and sharedkey and ttl and scopes:
                self.auth = JWTAuth(appid, clientid, sharedkey, ttl, scopes)
            else:
                raise ValueError(
                    "applicationId, clientId, sharedKey, token TTL and list of scopes must be provided "
                    "for JWT authentification"
                )
        else:
            raise Exception(f"Incorrect auth method {auth} specified")

        self.url = url
        self.api_md = "%s/md/{}" % self.url
        self.api_trade = "%s/trade/{}" % self.url
        self.session = requests.Session()
        self.session.mount(self.url, adapters.HTTPAdapter())
        self.logger = logger or logging.getLogger("HTTPApi")
        self.logger.setLevel(log_level)

    @backoff.on_exception(
        backoff.constant,
        (
            exceptions.ConnectionError,
            exceptions.Timeout,
            exceptions.ConnectTimeout,
            exceptions.ReadTimeout,
        ),
        max_tries=5,
        max_time=60,
    )
    def __request(
        self,
        method: Callable,
        api: str,
        handler: str,
        version: str,
        params: Optional[Dict[str, Any]] = None,
        jdata: Optional[Dict[str, Any]] = None,
        secret_jdata: Optional[Dict[str, Any]] = None,
    ) -> requests.Response:
        """
        wrapper for requests
        :param method: self.session.get or self.session.post
        :param api: self.api_feed or self.api_trade addresses
        :param handler: API handle
        :param params: additional params
        :param jdata: data to be passed as JSON
        :return: requests response object
        """
        headers = {"Accept": "application/json", "Accept-Encoding": "gzip"}
        self.logger.debug("received url %s" % api.format(version) + handler)
        self.logger.debug("passed headers: %s" % headers)
        if params:
            self.logger.debug("passed params: %s" % params)
        if jdata:
            self.logger.debug("passed json data: %s" % jdata)
        if secret_jdata and jdata:
            jdata.update(secret_jdata)
        elif secret_jdata:
            jdata = secret_jdata
        r = method(
            api.format(version) + handler,
            params=params,
            headers=headers,
            json=jdata,
            auth=self.auth,
        )
        if r.status_code > 204:
            self.logger.error("Got [%d] response" % r.status_code)
        else:
            self.logger.info("Got [%d] response" % r.status_code)
        self.logger.debug(r.text)
        return r

    @staticmethod
    def _mk_account(
        account: Optional[str], version: str = None
    ) -> Dict[str, Optional[str]]:
        if version == "3.0":
            return {"accountId": account}
        else:
            return {"account": account}

    @staticmethod
    def _mk_symbol(symbol: str, version: str = None) -> Dict[str, Optional[str]]:
        if version == "3.0":
            return {"symbolId": resolve_symbol(symbol)}
        else:
            return {"instrument": resolve_symbol(symbol)}

    def _wrap(
        self, r: requests.Response, non_json_resp: bool
    ) -> Union[None, str, List, Dict]:
        """
        simple wrapper to catch JSON parser errors
        """
        try:
            return json.loads(r.text)
        except json.JSONDecodeError:
            if not non_json_resp:
                self.logger.warning("Unable to parse JSON from %s" % r.text)

    def _get(
        self,
        api: str,
        handler: str,
        version: str,
        params: Optional[Dict[str, Any]] = None,
        non_json_resp: bool = False,
    ) -> Union[None, Dict[str, Any], List[Dict[str, Any]]]:
        return self._wrap(
            self.__request(
                method=self.session.get,
                api=api,
                handler=handler,
                version=version or self.version,
                params=params,
            ),
            non_json_resp,
        )

    def _post(
        self,
        api: str,
        handler: str,
        version: str,
        params: Optional[Dict[str, Any]] = None,
        jdata: Optional[Dict[str, Any]] = None,
        secret_jdata: Optional[Dict[str, Any]] = None,
        non_json_resp: bool = False,
    ) -> Union[None, str, Dict[str, Any], List[Dict[str, Any]]]:
        return self._wrap(
            self.__request(
                method=self.session.post,
                api=api,
                handler=handler,
                version=version or self.version,
                params=params,
                jdata=jdata,
                secret_jdata=secret_jdata,
            ),
            non_json_resp,
        )

    def _delete(
        self,
        api: str,
        handler: str,
        version: str,
        params: Optional[Dict[str, Any]] = None,
        non_json_resp: bool = False,
    ) -> Union[None, str, Dict[str, Any], List[Dict[str, Any]]]:
        return self._wrap(
            self.__request(
                method=self.session.delete,
                api=api,
                handler=handler,
                version=version or self.version,
                params=params,
            ),
            non_json_resp,
        )

    def get_user_accounts(self, version: str = None) -> Optional[List[UserAccount]]:
        """
        Return the list of user accounts and their statuses
        :param version: any version
        :return: UserAccount object
        """
        return extract_to_model(
            self._get(self.api_md, "/accounts", version), UserAccount, self._eraise
        )

    def get_changes(
        self, symbolid: Union[str, Iterable[str], None] = None, version: str = None
    ) -> Optional[List[ChangeType]]:
        """
        Return the list of daily changes for all or requested instruments
        :param symbolid: single string or list of instruments
        :param version: any version
        :return: List of ChangeType objects of specified versions
        """
        model = resolve_model(version or self.version, ChangeType)
        if symbolid:
            return extract_to_model(
                self._get(
                    self.api_md,
                    "/change/{}".format(
                        urlencode(symbolid, safe="")
                        if isinstance(symbolid, str)
                        else urlencode(",".join(symbolid), safe=",")
                    ),
                    version,
                ),
                model,
                self._eraise,
            )
        else:
            return extract_to_model(
                self._get(self.api_md, "/change", version), model, self._eraise
            )

    def get_currencies(self, version: str = None) -> List[str]:
        """
        Return the list of available currencies
        :param version: all versions
        :return: List of currencies
        """
        return self._get(self.api_md, "/crossrates", version).get("currencies", [])

    def get_crossrates(
        self, fr: str = "EUR", to: str = "USD", version: str = None
    ) -> Optional[Crossrate]:
        """
        Return the crossrate from one currency to another
        :param fr: pair from
        :param to: pair to
        :param version: all versions
        :return:
        """
        return extract_to_model(
            self._get(self.api_md, f"/crossrates/{fr.upper()}/{to.upper()}", version),
            Crossrate,
            self._eraise,
        )

    def get_exchanges(self, version: str = None) -> Optional[List[Exchange]]:
        """
        Return list of exchanges
        :param version: any version
        :return: List of Exchanges
        """
        return extract_to_model(
            self._get(self.api_md, "/exchanges", version), Exchange, self._eraise
        )

    def get_symbols_by_exch(
        self, exchange: Union[str, Exchange], version: str = None
    ) -> Optional[List[SymbolType]]:
        """
        Return the requested exchange financial instruments
        :param exchange: exchange
        :param version: all versions
        :return: List of Symbols
        """
        model = resolve_model(version or self.version, SymbolType)
        return extract_to_model(
            self._get(self.api_md, f"/exchanges/{attr_or(exchange, 'id_')}", version),
            model,
            self._eraise,
        )

    def get_groups(self, version: str = None) -> Optional[List[Group]]:
        """
        Return list of available instrument groups
        :param version: all versions
        :return: List of Groups
        """
        return extract_to_model(
            self._get(self.api_md, "/groups", version), Group, self._eraise
        )

    def get_symbols_by_gr(
        self, group: Union[str, Group], version: str = None
    ) -> Optional[List[SymbolType]]:
        """
        Return financial instruments which belong to specified group
        :param group: group string to Group object
        :param version: all versions
        :return: List of Symbols
        """
        model = resolve_model(version or self.version, SymbolType)
        return extract_to_model(
            self._get(self.api_md, f"/groups/{attr_or(group, 'group')}", version),
            model,
            self._eraise,
        )

    def get_nearest(
        self, group: Union[str, Group], version: str = None
    ) -> Optional[SymbolType]:
        """
        Return financial instrument which has the nearest expiration in the group
        :param group: group string to Group object
        :param version: 1.0 or 2.0
        :return: Symbol
        """
        if (version or self.version) not in (
            "1.0",
            "2.0",
        ):
            raise NotImplementedError(f"API not available in {version or self.version}")
        model = resolve_model(version or self.version, SymbolType)
        return extract_to_model(
            self._get(
                self.api_md, f"/groups/{attr_or(group, 'group')}/nearest", version
            ),
            model,
            self._eraise,
        )

    def get_symbols(self, version: str = None) -> Optional[List[SymbolType]]:
        """
        Return list of instruments available for authorized user
        :param version: all versions
        :return: List of Symbols
        """
        model = resolve_model(version or self.version, SymbolType)
        return extract_to_model(
            self._get(self.api_md, "/symbols", version), model, self._eraise
        )

    def get_symbol(self, symbol: str, version: str = None) -> Optional[SymbolType]:
        """
        Return instrument available for authorized user
        :param symbol: symbolId
        :param version: all versions
        :return: Symbol
        """
        model = resolve_model(version or self.version, SymbolType)
        return extract_to_model(
            self._get(self.api_md, f"/symbols/{symbol}", version), model, self._eraise
        )

    def get_symbol_schedule(
        self, symbol: Union[str, SymbolType], types: bool = False, version: str = None
    ) -> Optional[Schedule]:
        """
        Return financial schedule for requested instrument
        :param symbol: SymbolType or symbol string
        :param types: show order types in response
        :param version: all versions
        :return: Schedule
        """
        return extract_to_model(
            self._get(
                self.api_md,
                f"/symbols/{resolve_symbol(symbol)}/schedule",
                version,
                params={"types": str(types).lower()},
            ),
            Schedule,
            self._eraise,
        )

    def get_symbol_spec(
        self, symbol: Union[str, SymbolType], version: str = None
    ) -> Optional[SymbolSpecification]:
        """
        Return additional parameters for requested instrument
        :param symbol: SymbolType or symbol string
        :param version: all versions
        :return: SymbolSpecification
        """
        return extract_to_model(
            self._get(
                self.api_md, f"/symbols/{resolve_symbol(symbol)}/specification", version
            ),
            SymbolSpecification,
            self._eraise,
        )

    def get_types(self, version: str = None) -> List[InstrumentType]:
        """
        Return list of known instrument types
        :param version: all versions
        :return: List of InstrumentTypes
        """
        return [
            InstrumentType(x["id"]) for x in self._get(self.api_md, "/types", version)
        ]

    def get_symbol_by_type(
        self, sym_type: InstrumentType, version: str = None
    ) -> Optional[List[SymbolType]]:
        """
        Return financial instruments of the requrested type
        :param sym_type: InstrumentType Enum
        :param version: all versions
        :return: List of Symbols
        """
        model = resolve_model(version or self.version, SymbolType)
        return extract_to_model(
            self._get(self.api_md, f"/types/{sym_type.value}", version),
            model,
            self._eraise,
        )

    def get_quote_stream(
        self,
        symbols: Union[str, SymbolType, Iterable[str], Iterable[SymbolType]],
        level: FeedLevel = FeedLevel.BP,
        version: str = None,
    ) -> HTTPApiStreaming:
        """
        Return the life quote stream for the specified financial instruments
        :param symbols: single or List/Tuple/etc of Symbols
        :param level: Quote Type - FeedLevel enum: BP|BEST_PRICE for top book and MD|MARKET_DEPTH for full market
        :param version: all versions
        :return: HTTPApiStreaming instance thread
        """
        return HTTPApiStreaming(
            self.auth,
            self.api_md.format(version or self.version),
            f"/feed/{resolve_symbol(symbols)}",
            resolve_model(version or self.version, QuoteType),
            {"level": level.value},
            logger=self.logger,
            log_level=self.logger.getEffectiveLevel(),
            debug_raise_model_exceptions=self._eraise,
        )

    def get_trade_stream(
        self,
        symbols: Union[str, SymbolType, Iterable[str], Iterable[SymbolType]],
        version: str = None,
    ) -> HTTPApiStreaming:
        """
        Return the trades stream for the specified financial instruments
        :param symbols: single or List/Tuple/etc of Symbols
        :param version: 3.0 version
        :return: HTTPApiStreaming instance thread
        """
        if (version or self.version) not in ("3.0",):
            raise NotImplementedError(f"API not available in {version or self.version}")
        return HTTPApiStreaming(
            self.auth,
            self.api_md.format(version or self.version),
            f"/feed/trades/{resolve_symbol(symbols)}",
            resolve_model(version or self.version, TradeType),
            logger=self.logger,
            log_level=self.logger.getEffectiveLevel(),
            debug_raise_model_exceptions=self._eraise,
        )

    def get_last_quote(
        self,
        symbol: Union[str, Iterable[str], SymbolType, Iterable[SymbolType]],
        level: FeedLevel = FeedLevel.BP,
        version: str = None,
    ) -> Optional[List[QuoteType]]:
        """
        Return the last quote for the specified financial instrument
        :param symbol: symbol string or Symbol object
        :param level: Quote Type - FeedLevel enum: BP|BEST_PRICE for top book and MD|MARKET_DEPTH for full market
        :param version: all versions, MD level available since 2.0
        :return: List of Quotes
        """
        model = resolve_model(version or self.version, QuoteType)
        return extract_to_model(
            self._get(
                self.api_md,
                f"/feed/{resolve_symbol(symbol)}/last",
                version,
                {"level": level.value},
            ),
            model,
            self._eraise,
        )

    def get_ohlc(
        self,
        symbol: Union[str, SymbolType],
        duration: Union[int, CandleDurations],
        agg_type: DataType = DataType.QUOTES,
        start: Optional[Union[Numeric, datetime]] = None,
        end: Optional[Union[Numeric, datetime]] = None,
        limit: int = 60,
        version: str = None,
    ) -> Optional[List[Union[OHLCQuotes, OHLCTrades]]]:
        """
        Return the list of OHLC candles for the specified financial instrument and duration
        :param symbol: symbol string or Symbol object
        :param duration: aggregate intreval in seconds or one of CandleDurations Enum
        :param agg_type: DataType Enum
        :param start: starting timestamp(ms) or datetime object in UTC timezone
        :param end: ending timestamp(ms) or datetime object in UTC timezone
        :param limit: maximum candle amount retrieved. Can be reduced by server
        :param version: all versions
        :return: List of OHLC objects
        """
        s = resolve_symbol(symbol)
        params = {
            "size": limit,
            "type": agg_type.value,
            "from": dt_to_timestamp(start, True)
            if isinstance(start, datetime)
            else opt_int(start),
            "to": dt_to_timestamp(end, True)
            if isinstance(end, datetime)
            else opt_int(end),
        }
        return extract_to_model(
            self._get(
                self.api_md, f"/ohlc/{s}/{attr_or(duration, 'value')}", version, params
            ),
            OHLCQuotes if agg_type == DataType.QUOTES else OHLCTrades,
            self._eraise,
        )

    def get_ticks(
        self,
        symbol: Union[str, SymbolType],
        agg_type: DataType = DataType.QUOTES,
        start: Optional[Union[Numeric, datetime]] = None,
        end: Optional[Union[Numeric, datetime]] = None,
        limit: int = 1000,
        version: str = None,
    ) -> Optional[List[Union[QuoteType, TradeType]]]:
        """
        Return the list of ticks for the specified financial instrument
        :param symbol: symbol string or Symbol object
        :param agg_type: DataType Enum
        :param start: starting timestamp(ms) or datetime object in UTC timezone
        :param end: ending timestamp(ms) or datetime object in UTC timezone
        :param limit: maximum candle amount retrieved. Can be reduced by server
        :param version: available since 2.0
        :return: List of Quotes or Trades
        """
        if agg_type == DataType.QUOTES:
            model = resolve_model(version or self.version, QuoteType)
        else:
            model = resolve_model(version or self.version, TradeType)
        params = {
            "size": limit,
            "type": agg_type.value,
            "from": dt_to_timestamp(start, True)
            if isinstance(start, datetime)
            else opt_int(start),
            "to": dt_to_timestamp(end, True)
            if isinstance(end, datetime)
            else opt_int(end),
        }
        return extract_to_model(
            self._get(self.api_md, f"/ticks/{resolve_symbol(symbol)}", version, params),
            model,
            self._eraise,
        )

    def get_account_summary(
        self,
        account: str,
        currency: str = "EUR",
        req_date: Union[str, datetime] = None,
        version: str = None,
    ) -> Optional[SummaryType]:
        """
        Return the summary for the specified account
        :param account: accountId
        :param currency: on of NAV currency, default is EUR
        :param req_date: historical account summary, datetime or YYYY-MM-DD string
        :param version: all versions
        :return: AccountSummary
        """
        model = resolve_model(version or self.version, SummaryType)
        if req_date:
            if isinstance(req_date, datetime):
                d = dt_to_str(req_date, "%Y-%m-%d")
            else:
                d = req_date
            return extract_to_model(
                self._get(
                    self.api_md, f"/summary/{account}/{d}/{currency.upper()}", version
                ),
                model,
                self._eraise,
            )
        else:
            return extract_to_model(
                self._get(
                    self.api_md, f"/summary/{account}/{currency.upper()}", version
                ),
                model,
                self._eraise,
            )

    def get_transactions(
        self,
        account: Optional[str] = None,
        uuid: Optional[str] = None,
        symbol: Union[str, SymbolType, None] = None,
        asset: Optional[str] = None,
        op_type: Union[str, Iterable[str], None] = None,
        order_id: Optional[str] = None,
        order_pos: Optional[int] = None,
        offset: Numeric = None,
        limit: Numeric = 10,
        order: Ordering = Ordering.ASC,
        fr: Union[int, datetime] = None,
        to: Union[int, datetime] = None,
        version: str = None,
    ) -> Optional[List[TransactionType]]:
        """
        Return the list of transactions with the specified filter
        :param account: filter transactions by accountId [single account]
        :param uuid: filter transactions by UUID [single uuid]
        :param symbol: filter transactions by symbol [single symol]
        :param asset: filter transactions by asset [single asset]
        :param op_type: filter transactions by operationType - note, this list is subject to change.
        It's not recommended to Enum this
        :param order_id: filter transactions by order UUID [single order]
        :param order_pos: filter transactions by order position
        :param offset: offset of retrived transaction list. Can be used for pagination
        :param limit: limit number of retrieved transactions. Can be maxed on server side
        :param order: Ordering enum: ASC or DESC ordering
        :param fr: retrive transactions from date - timestamp in ms or datetime (UTC timezone)
        :param to: retrive transactions till date - timestamp in ms or datetime (UTC timezone)
        :param version: all version
        :return: List of Transactions
        """
        params = {
            "uuid": uuid,
            "accountId": account,
            "symbolId": resolve_symbol(symbol),
            "asset": asset,
            "offset": opt_int(offset),
            "limit": opt_int(limit),
            "order": order.value,
            "orderId": order_id,
            "orderPos": opt_int(order_pos),
            "fromDate": dt_to_str(fr, "%Y-%m-%d")
            if isinstance(fr, datetime)
            else dt_to_str(timestamp_to_dt(fr), "%Y-%m-%d"),
            "toDate": dt_to_str(to, "%Y-%m-%d")
            if isinstance(to, datetime)
            else dt_to_str(timestamp_to_dt(to), "%Y-%m-%d"),
            # full datetime filter doesn't work
            # "fromDate": dt_to_str(fr) if isinstance(fr, datetime) else dt_to_str(timestamp_to_dt(fr)),
            # "toDate": dt_to_str(to) if isinstance(to, datetime) else dt_to_str(timestamp_to_dt(to))
        }
        if hasattr(op_type, "__iter__") and not isinstance(op_type, str):
            params.update({"operationType": ",".join(op_type)})
        else:
            params.update({"operationType": op_type})
        model = resolve_model(version or self.version, TransactionType)
        return extract_to_model(
            self._get(self.api_md, "/transactions", version, params),
            model,
            self._eraise,
        )

    def place_order(
        self, order: OrderSentType, version: str = None
    ) -> Optional[List[Union[OrderType, Reject]]]:
        """
        Place new trading order
        :param order: Order Model
        :param version: all versions, check the differences in naming models
        :return: List of Orders
        """
        model = resolve_model(version or self.version, OrderType)
        r = self._post(self.api_trade, "/orders", version, jdata=order.to_json())
        if version == "1.0":
            return [extract_to_model(r, model, self._eraise, backup_obj=Reject)]
        else:
            return extract_to_model(r, model, self._eraise, backup_obj=Reject)

    def get_orders(
        self,
        account: Optional[str] = None,
        limit: Numeric = 1000,
        fr: Union[int, datetime] = None,
        to: Union[int, datetime] = None,
        version: str = None,
    ) -> Optional[List[OrderType]]:
        """
        Return the list of historical orders
        :param account: account permissioned to request
        :param limit: max size of response
        :param fr: retrieve orders placed from date - timestamp in ms or datetime (UTC timezone)
        :param to: retrieve order placed before date - timestamp in ms or datetime (UTC timezone)
        :param version: all versions
        :return: List of Orders
        """
        model = resolve_model(version or self.version, OrderType)
        params = {
            "limit": opt_int(limit),
            "from": dt_to_str(fr)
            if isinstance(fr, datetime)
            else dt_to_str(timestamp_to_dt(fr)),
            "to": dt_to_str(to)
            if isinstance(to, datetime)
            else dt_to_str(timestamp_to_dt(to)),
        }
        if account:
            params.update(HTTPApi._mk_account(account, version or self.version))
        return extract_to_model(
            self._get(self.api_trade, "/orders", version, params), model, self._eraise
        )

    def get_active_orders(
        self,
        account: Optional[str] = None,
        limit: Numeric = 10,
        symbol: Union[str, SymbolType, None] = None,
        version: str = None,
    ) -> Optional[List[OrderType]]:
        """
        Return the list of active trading orders
        :param account: account permissioned to request
        :param limit: max size of response
        :param symbol: filter orders by symbol
        :param version: all versions
        :return: List of Orders
        """
        params = {"limit": opt_int(limit)}
        params.update(HTTPApi._mk_account(account, version))
        params.update(HTTPApi._mk_symbol(symbol, version))
        model = resolve_model(version or self.version, OrderType)
        return extract_to_model(
            self._get(self.api_trade, "/orders/active", version, params),
            model,
            self._eraise,
        )

    def get_order(
        self, orderid: Union[str, OrderType], version: str = None
    ) -> Optional[OrderType]:
        """
        Return the order with specified identifier
        :param orderid: OrderId via string or Order object
        :param version: all versions
        :return: Order
        """
        model = resolve_model(version or self.version, OrderType)
        if isinstance(orderid, OrderV3):
            id_ = orderid.order_id
        elif isinstance(orderid, (OrderV2, OrderV1)):
            id_ = orderid.id_
        else:
            id_ = orderid
        return extract_to_model(
            self._get(self.api_trade, f"/orders/{id_}", version), model, self._eraise
        )

    def _modify_order(
        self,
        orderid: Union[str, OrderType],
        action: ModifyAction,
        version: str = None,
        **kwargs,
    ) -> Union[None, Reject, OrderType]:
        """
        Replace or cancel trading order
        :param orderid: orderId
        :param action: replace or cancel
        :param version: 1.0 or 2.0. reffer docs to see difference
        :param kwargs: quantity and/or limitPrice for limit or stop-limit, stopPrice for stop or stop-limit order
        :return: dict
        """
        model = resolve_model(version or self.version, OrderType)
        data = {"action": action.value}
        if action == ModifyAction.REPLACE:
            data["parameters"] = {}
            if kwargs:
                for k, v in kwargs.items():
                    if v is not None:
                        data["parameters"][k] = str(v)
        r = self._post(self.api_trade, f"/orders/{orderid}", version, jdata=data)
        return extract_to_model(r, model, self._eraise, backup_obj=Reject)

    def cancel_order(
        self, orderid: Union[str, OrderType], version: str = None
    ) -> Union[None, Reject, OrderType]:
        """
        Cancel trading order
        :param orderid: orderId
        :param version: all versions
        :return: OrderType or Reject
        """
        return self._modify_order(orderid, ModifyAction.CANCEL, version)

    def replace_order(
        self,
        orderid: Union[str, OrderType],
        quantity: Numeric,
        limit_price: Optional[Numeric] = None,
        stop_price: Optional[Numeric] = None,
        price_distance: Optional[Numeric] = None,
        version: str = None,
    ) -> Union[None, Reject, OrderType]:
        """
        Replace trading order
        :param orderid: orderId
        :param quantity: original or changed quantity for order
        :param limit_price: applied for Limit, StopLimit and Iceberg type of orders
        :param stop_price: applied for Stop and StopLimit type of orders
        :param price_distance: applied for TrailingStop type of orders
        :param version: all versions
        :return: OrderType or List of Rejects
        """
        return self._modify_order(
            orderid,
            ModifyAction.REPLACE,
            version,
            quantity=quantity,
            limitPrice=limit_price,
            stopPrice=stop_price,
            priceDistance=price_distance,
        )

    def get_orders_stream(self, version: str = None) -> HTTPApiStreaming:
        """
        Return the life quote stream for the specified financial instruments
        :param version: all versions
        :return: HTTPApiStreaming instance thread
        """
        return HTTPApiStreaming(
            self.auth,
            self.api_trade.format(version or self.version),
            "/stream/orders",
            resolve_model(version or self.version, OrderType),
            event_filter="order",
            logger=self.logger,
            log_level=self.logger.getEffectiveLevel(),
            debug_raise_model_exceptions=self._eraise,
        )

    def get_exec_orders_stream(self, version: str = None) -> HTTPApiStreaming:
        """
        Return the life quote stream for the specified financial instruments
        :param version: all versions
        :return: HTTPApiStreaming instance thread
        """
        return HTTPApiStreaming(
            self.auth,
            self.api_trade.format(version or self.version),
            "/stream/trades",
            resolve_model(version or self.version, ExOrderType),
            event_filter="trade" if (version or self.version) == "3.0" else None,
            logger=self.logger,
            log_level=self.logger.getEffectiveLevel(),
            debug_raise_model_exceptions=self._eraise,
        )



from backtesterrb30.libs.utils.run_service import run_service

from backtesterrb30.live_data_feeds.live_data_feeds import LiveDataFeeds

microservice_name = "live_data_feeds"

run_service(microservice_name, LiveDataFeeds)



import os

from backtesterrb30.libs.utils.module_loaders import import_executor_module
from backtesterrb30.libs.utils.run_service import run_service

here = os.getcwd()
strategy_path = os.getenv("STRATEGY_PATH")
strategy_file = os.getenv("STRATEGY_FILE")
executor_class_name = os.getenv("EXECUTOR_CLASS_NAME")


microservice_name = "python_executor"
module = import_executor_module(strategy_path, strategy_file, executor_class_name)
run_service(microservice_name, module)



import asyncio
from enum import Enum
from os import getenv
from time import time
from typing import Union

import pandas as pd

from backtesterrb30.libs.data_sources.data_source_base import DataSource
from backtesterrb30.libs.data_sources.utils import validate_dataframe_timestamps
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol

# from backtesterrb30.libs.utils.historical_sources import EXANTE_INTERVALS
from backtesterrb30.libs.xnt.http_api import (
    AuthMethods,
    CandleDurations,
    DataType,
    HTTPApi,
)


class EXANTE_INTERVALS_2(str, Enum):
    tick: str = "tick"
    minute: str = "minute"
    minute5: str = "minute5"
    minute30: str = "minute30"
    hour: str = "hour"
    day: str = "day"


class ExanteDataSource(DataSource):
    INTERVALS = EXANTE_INTERVALS_2
    NAME = "exante"

    def __init__(self, logger=print):
        super().__init__(True, logger)
        exante_app_id = getenv("exante_app_id")
        exante_access_key = getenv("exante_access_key")
        if exante_app_id is None or exante_access_key is None:
            raise Exception(self.NAME + " No authorization heys provided")
        self.client = HTTPApi(
            AuthMethods.BASIC, appid=exante_app_id, acckey=exante_access_key
        )

    def __get_exante_interval(self, interval: str) -> CandleDurations:
        if interval == "minute":
            return CandleDurations.MIN1
        if interval == "minute5":
            return CandleDurations.MIN5
        if interval == "minute30":
            return CandleDurations.MIN30
        if interval == "hour":
            return CandleDurations.HOUR1
        if interval == "day":
            return CandleDurations.DAY1

    def _get_interval_miliseconds(self, interval: str) -> Union[int, None]:
        return int(self.__get_exante_interval(interval).value * 1000)

    # override
    async def _validate_instrument_data(self, data: DataSymbol) -> bool:
        # TODO check if volume necessery or not.

        data_type = DataType.QUOTES
        if data.with_volume:
            data_type = DataType.TRADES

        start_validation_time = time()
        candles = None
        while True:
            candles = self.client.get_ohlc(
                symbol=data.symbol,
                duration=60,
                start=1,
                version="3.0",
                limit=1,
                agg_type=data_type,
            )
            if candles is not None or time() - start_validation_time > 61:
                break
            self._log(
                "Performing Exante validation for this instrument going to take up to 1 minute"
            )
            await asyncio.sleep(10)
        if candles is None:
            self._log(
                'Error. Instrument "'
                + data.symbol
                + '" probably does not exists on exante with this configuration. Try without volume'
            )
            return False
        # from_datetime_timestamp = int(from_datetime.timestamp() * 1000)
        first_datetime = candles[0].timestamp
        if first_datetime > data.backtest_date_start:
            self._log(
                "Error. First avaliable date of ", data.symbol, "is", first_datetime
            )
            return False
        return True

    async def __wait(self):
        self._log("waiting for exante download 60s")
        await asyncio.sleep(60)

    # override
    async def _download_instrument_data(
        self,
        instrument: str,
        interval: str,
        time_start: int,
        time_stop: Union[int, None],
    ) -> pd.DataFrame:
        self._log("Downloading exante data", instrument)
        # try:
        await self.__wait()
        if interval == "tick":
            limit = 5000
            floating_start = time_start
            ticks_arr = []
            iteration_n = 0
            while True:
                if floating_start >= time_stop:
                    break
                ticks = self.client.get_ticks(
                    symbol=instrument,
                    start=floating_start,
                    end=time_stop,
                    limit=limit,
                    agg_type=DataType.QUOTES,
                )
                if ticks is None:
                    self._raise_error("Error while downloading ticks")
                ticks_arr = ticks + ticks_arr
                if ticks == []:
                    if iteration_n == 0:
                        self._raise_error("Error while downloading ticks ...")
                if len(ticks) == 5000:
                    floating_start = int(ticks[0].timestamp.timestamp() * 1000) + 1

                if len(ticks) < 5000:
                    break
                iteration_n += 1
                await self.__wait()

            ticks_transformed = []
            for tck in ticks_arr:
                ticks_transformed.append(
                    [int(tck.timestamp.timestamp() * 1000), tck.ask[0].price]
                )
            df = pd.DataFrame(ticks_transformed)
            df = df.iloc[::-1]

        else:
            limit = 5000
            exante_interval = self.__get_exante_interval(interval).value
            floating_start = time_start
            klines_arr = []
            iteration_n = 0
            while True:
                if floating_start >= time_stop:
                    break
                klines = self.client.get_ohlc(
                    symbol=instrument,
                    duration=exante_interval,
                    start=floating_start,
                    end=time_stop,
                    limit=limit,
                    agg_type=DataType.QUOTES,
                )
                if klines is None:
                    self._raise_error("Error while downloading klines")
                klines_arr = klines + klines_arr
                if klines == []:
                    if iteration_n == 0:
                        self._raise_error("Error while downloading klines ...")
                if len(klines) == 5000:
                    floating_start = int(klines[0].timestamp.timestamp() * 1000) + (
                        exante_interval * 1000
                    )

                if len(klines) < 5000:
                    break
                iteration_n += 1
                await self.__wait()

            # klines_transformed = []
            # for kl in klines_arr:
            #     # klines_transformed.append([int(round(datetime.timestamp(kl.timestamp) * 1000)), kl.open_])
            #     klines_transformed.append([int(kl.timestamp.timestamp() * 1000), kl.open_])
            # df = pd.DataFrame(klines_transformed)
            # df = df.iloc[::-1]

            klines_transformed = []
            exante_interval = self.__get_exante_interval(interval).value
            prev_kl = None
            for kl in reversed(klines_arr):
                if (
                    prev_kl
                    and kl.timestamp.timestamp() * 1000
                    - prev_kl.timestamp.timestamp() * 1000
                    >= exante_interval * 1000 * 2
                ):
                    new_timestamp = int(prev_kl.timestamp.timestamp() * 1000) + (
                        exante_interval * 1000
                    )
                    klines_transformed.append([new_timestamp, prev_kl.close])
                klines_transformed.append(
                    [int(kl.timestamp.timestamp() * 1000), kl.open_]
                )
                prev_kl = kl
            df = pd.DataFrame(klines_transformed)
            # self._log('exante shape before validation', df.shape)
            df = validate_dataframe_timestamps(
                df,
                self.__get_exante_interval(interval).value * 1000,
                time_start,
                time_stop,
            )
            # self._log('exante shape after validation', df.shape)
            # print('head', str(df.head(1)))
            # print('tail', str(df.tail(1)))

        return df



from abc import ABC, abstractmethod
from asyncio import AbstractEventLoop
from typing import Callable

from backtesterrb30.libs.utils.list_of_services import SERVICES


class BrokerBase(ABC):
    @abstractmethod
    def run(self):
        pass

    @abstractmethod
    def send(self, service: SERVICES, msg: str, *args):
        pass

    @abstractmethod
    def register(self, command: str, func: Callable):
        pass

    @abstractmethod
    def create_listeners(self, loop: AbstractEventLoop):
        pass

    @abstractmethod
    def stop(self):
        pass

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def pause(self):
        pass



import os

from backtesterrb30.libs.utils.module_loaders import import_model_module
from backtesterrb30.libs.utils.run_service import run_service

strategy_path = os.getenv("STRATEGY_PATH")
strategy_file = os.getenv("STRATEGY_FILE")
model_class_name = os.getenv("MODEL_CLASS_NAME")


microservice_name = "python_engine"
module = import_model_module(strategy_path, strategy_file, model_class_name)
run_service(microservice_name, module)



#!/usr/bin/env python3.7
import json
import warnings
from datetime import datetime, timezone
from decimal import Decimal
from enum import Enum, EnumMeta
from inspect import signature
from typing import Any, Callable, Dict, List, Optional, Set, Type, TypeVar, Union

from deepdiff import DeepDiff
from inflection import camelize, underscore

reserved = ("type", "id", "list", "except", "from", "to", "open", "sum", "uuid")
SerializableType = TypeVar("SerializableType", bound="Serializable")

Numeric = Union[int, float, str, Decimal]


def camel(s: str, uppercase_first_letter=False):
    if s.endswith("_"):
        return camelize(s[:-1], uppercase_first_letter)
    else:
        return camelize(s, uppercase_first_letter)


def extract_to_model(
    data: Any,
    obj: Type[SerializableType],
    eraise: bool = False,
    backup_obj: Optional[Type[SerializableType]] = None,
) -> Union[None, SerializableType, List[SerializableType]]:
    """
    Serialize JSON-like data to Serializable object, tries backup_obj if supplied
    :param data: incoming data, list or dicts
    :param obj: child of Serializable class
    :param eraise: raise RuntimeError if incorrect data supplied or model out-of-data
    :param backup_obj: trying to serialize data into backup_obj if not successfull on obj
    :return: Serializable class instance if possible
    """
    if isinstance(data, Dict):
        if hasattr(obj, "__model__"):
            r = obj.from_json(data)
        elif callable(obj):
            r = obj(data)
        else:
            # this should never happens
            raise ValueError(f"Object {obj} has undeterminable type {type(obj)}")
        if isinstance(r, RuntimeError) and backup_obj is not None:
            return extract_to_model(data, backup_obj, True)
        elif isinstance(r, RuntimeError) and not backup_obj:
            if eraise:
                raise r
            else:
                warnings.warn("Unable to wrap data '%s', skipping" % data)
                return None
        elif not isinstance(r, Exception):
            return r
        else:
            # this should never happens, added for lintering
            raise r
    elif isinstance(data, (List, Set)):
        return [extract_to_model(x, obj, eraise, backup_obj) for x in data]
    elif isinstance(data, obj):
        return data
    else:
        return None


def to_string(d: datetime, fmt: str = "%Y-%m-%dT%H:%M:%S.%fZ") -> str:
    return d.strftime(fmt)


def dt_to_timestamp(d: datetime, millis: bool = False) -> int:
    if millis:
        return int(d.timestamp() * 1000)
    else:
        return int(d.timestamp())


def timestamp_to_dt(
    ts: Optional[Numeric], tz: timezone = timezone.utc
) -> Optional[datetime]:
    if ts is None:
        return None
    else:
        try:
            if isinstance(ts, (str, Decimal)):
                return datetime.fromtimestamp(float(ts), tz=tz)
            else:
                return datetime.fromtimestamp(ts, tz=tz)
        except ValueError:
            return timestamp_to_dt(float(ts) / 1000, tz)


def str_to_dt(
    s: Optional[str], fmt: str = "%Y-%m-%dT%H:%M:%S.%f%z"
) -> Optional[datetime]:
    try:
        return datetime.strptime(s, fmt)
    except ValueError:
        return None


def dt_to_str(
    d: Optional[datetime], fmt: str = "%Y-%m-%dT%H:%M:%S.%f%z"
) -> Optional[str]:
    if d:
        return d.strftime(fmt)
    else:
        return None


def dc(value: Optional[Numeric]) -> Optional[Decimal]:
    if value is None:
        return None
    elif isinstance(value, float):
        return Decimal(str(value))
    else:
        return Decimal(value)


def opt_int(i: Optional[Numeric]) -> Optional[int]:
    if i is None:
        return None
    else:
        return int(dc(i))


def attr_or(obj: Any, attribute: str) -> Optional[Any]:
    if obj is None:
        return None
    if not hasattr(obj, attribute):
        return obj
    else:
        return getattr(obj, attribute, None)


class BaseSerializable:
    __model__ = True

    def __dict(self, obj: Any, dt_parser: Callable, keep_null: bool) -> Any:
        if isinstance(obj, Dict):
            return {
                self.__dict(key, dt_parser, keep_null): self.__dict(
                    value, dt_parser, keep_null
                )
                for key, value in obj.items()
                if (not keep_null and value is not None) or keep_null
            }
        elif isinstance(obj, Decimal):
            return str(obj)
        elif isinstance(obj, Enum):
            return obj.value
        elif isinstance(obj, datetime):
            return dt_parser(obj)
        elif hasattr(obj, "__iter__") and not isinstance(obj, str):
            return [
                self.__dict(value, dt_parser, keep_null)
                for value in obj
                if (not keep_null and value is not None) or keep_null
            ]
        elif hasattr(obj, "__dict__"):
            data = {
                camel(key): self.__dict(value, dt_parser, keep_null)
                for key, value in obj.__dict__.items()
                if (not callable(value))
                and ((not keep_null and value is not None) or keep_null)
            }
            return data
        else:
            return obj

    @staticmethod
    def to_enum(
        source: Union[int, str, Enum, None], obj: EnumMeta, debug: bool = False
    ) -> Optional[Enum]:
        if source is None:
            return None
        elif isinstance(source, Enum):
            return source
        else:
            try:
                return obj(source)
            except ValueError:
                try:
                    return obj[source]
                except KeyError:
                    if debug:
                        return source
                    else:
                        raise ValueError(f"Unable to extract Enum from {obj}")

    def to_json(
        self, keep_null: bool = False, dt_parser: Callable = to_string
    ) -> Dict[str, Union[str, int, float]]:
        """
        Method to convert model to JSON
        :param keep_null: True to keep "field": null in generated JSON, default is False (no key)
        :param dt_parser: function to convert datetime object to printable
        :return: JSON-like dictionary
        """
        return self.__dict(self, dt_parser, keep_null)

    def __repr__(self) -> str:
        return json.dumps(self.to_json(False)).decode()

    def __eq__(self, other) -> bool:
        if hasattr(other, "__model__"):
            return bool(
                not DeepDiff(
                    self.to_json(keep_null=True),
                    other.to_json(keep_null=True),
                    ignore_order=True,
                )
            )
        else:
            raise ValueError("Impossible to directly compare non-models objects")


class Serializable(BaseSerializable):
    __model__ = True

    @classmethod
    def _empty_init(cls) -> Dict[str, None]:
        return {
            k: None
            for k in signature(cls.__init__).parameters.keys()
            if k not in ("args", "kwargs", "self")
        }

    @classmethod
    def from_json(
        cls: Type[SerializableType], obj: Union[Dict, SerializableType, None]
    ) -> Union[SerializableType, Exception]:
        def make_key(key: str) -> str:
            key = underscore(key)
            if key in reserved:
                return "%s_" % key
            else:
                return key

        try:
            if obj in (None, [], {}, [{}]):
                return cls(**cls._empty_init())  # type: ignore
            elif isinstance(obj, dict):
                data = {}  # type: Dict[str, Any]
                for k, v in obj.items():
                    if isinstance(v, (int, float, Decimal)) and not isinstance(v, bool):
                        data[make_key(k)] = dc(v)
                    else:
                        data[make_key(k)] = v
                return cls(**data)  # type: ignore
            elif isinstance(obj, cls):
                return obj
        except TypeError:
            return RuntimeError(f"could not get {cls} from {type(obj)} as {obj}")



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Callable


class Service(ABC):
    name: str
    __logger: Callable
    _broker: object

    def __init__(self, config, logger: Callable = print):
        self.name = config.name
        self.__logger = logger

    def run(self):
        if not self._broker:
            raise Exception("No communication broker registered")
        self._log("Running service")
        self._configure()
        self._broker.run()
        self._loop()

    def get_logger(self):
        return self._log

    @abstractmethod
    def _loop(self):
        pass

    # @abstractmethod
    # def _send(self, msg: dict, pub: dict):
    #     pass

    def _configure(self):
        pass

    def register_communication_broker(self, broker):
        self._broker = broker

    def _log(self, *msg):
        time = str(datetime.now())
        msg = " ".join(str(el) for el in msg)
        self.__logger(f"{time} [{self.name:>25}] {msg}")
        # self.__logger(f'{time}|  {msg}')



import asyncio
from abc import abstractmethod
from typing import List

from backtesterrb30.libs.communication_broker.broker_base import BrokerBase
from backtesterrb30.libs.interfaces.python_backtester.debug_breakpoint import (
    DebugBreakpoint,
)
from backtesterrb30.libs.interfaces.python_backtester.last_feed import LastFeed
from backtesterrb30.libs.interfaces.python_backtester.trade import Trade
from backtesterrb30.libs.interfaces.python_executor.executor_position import (
    ExecutorPosition,
)
from backtesterrb30.libs.interfaces.utils.config import Config
from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.utils.list_of_services import SERVICES
from backtesterrb30.libs.utils.service import Service


class Executor(Service):
    """Python Trade executor"""

    _broker: BrokerBase

    def __init__(
        self, config: Config, data_schema: DataSchema, loop=None, logger=print
    ):
        super().__init__(config, logger)
        self.config: Config = config
        self.__loop = loop
        self.__custom_event_loop = False
        if self.__loop is None:
            self.__loop = asyncio.get_event_loop()
            self.__custom_event_loop = True
        self.__data_schema: DataSchema = data_schema
        self.__columns = ["timestamp"] + [c.symbol for c in self.__data_schema.data]
        # self.__event_price = 0
        # self.__event_timestamp = 0
        # self.__event_last_feed = []
        # self.__number_of_actions = 0
        self.__current_capital = 0
        self.__current_invested = 0
        self.__start_amout_of_usd = 10000

        self.__positions: List[ExecutorPosition] = [
            {"instrument": "inst", "number_of_actions": 123, "current_value": 12}
        ]

    # public methods:
    # ==================================================================

    @abstractmethod
    async def on_event(self, message):
        """
        Function is being called when :class:`Model` class triggers
        `trigger_event()` function.

        :param message: message defined in `model.py` in strategy files.
        :type message: Any
        """
        self._log("method should be implemented in strategy function")

    def get_data_schema(self):
        """
        Returns data_schema defined in your strategy files in `data_schema.py`
        Data schema contains list of :class:`DataSymbol` objects. In every :class:`DataSymbol` object,
        you can use `get_buffer()` function to get current buffer connected to this symbol.

        :return: :class:`DataSchema` object combined with this strategy.
        :rtype: DataSchema
        """
        return self.__data_schema

    def get_data_symbol_by_custom_name(self, custom_name: str):
        """
        Returns a :class:`DataSymbol` object using `custom_name` defined
        in `data_schema.py`, `custom_name` must be unique.

        :param custom_name: custom name defined in `data_schema.py` file.
        :type custom_name: str
        :return: :class:`DataSymbol` object combined with privided custom_name
        :rtype: DataSymbol
        """
        if not isinstance(custom_name, str):
            raise Exception("Provided name is not string")
        arr = [d for d in self.__data_schema.data if d.custom_name == custom_name]
        if len(arr) == 0:
            raise Exception("No data symbol with such custom name")
        if len(arr) > 1:
            raise Exception("Two elements with the same custom name")
        return arr[0]

    async def trade(
        self,
        trade_value: float,
        data_symbol: DataSymbol,
        price: float = None,
        timestamp: int = None,
    ):
        """
        Triggers trade.

        :param trade_value: value of trade in dollars.
        :type trade_value: float
        :param data_symbol: :class:`DataSymbol` on which trade is going to be made.
        :type data_symbol: DataSymbol
        :param price: price of trade. Default is current price. Works only in backtest mode.
        :type price: float, optional
        :param timestamp: timestamp of trade. Default is current timestamp. Works only in backtest mode.
        :type timestamp: int, optional
        """
        if not price and not timestamp:
            price = 0
            timestamp = 0
        if price and not timestamp or timestamp and not price:
            raise Exception("Provide both price and timestamp or none of them.")
        if (
            trade_value
            > self.__start_amout_of_usd
            + self.__current_capital
            - self.__current_invested
        ):
            raise Exception("To big amout of trade")
        if self.config.backtest:
            trade_params = {
                "value": trade_value,
                "price": price,
                "timestamp": timestamp,
                "symbol": data_symbol.symbol,
                "source": data_symbol.historical_data_source,
            }
            await self._broker.send(
                SERVICES.python_backtester, "trade", Trade(**trade_params)
            )
            return True
        else:
            self._log("Live trading not implemented")

    async def close_all_trades(self):
        """Closes all opened trades."""
        if self.config.backtest:
            await self._broker.send(SERVICES.python_backtester, "close_all_trades")
        else:
            # TODO trade in real broker
            pass

    # def _get_number_of_actions(self):
    #     return self.__number_of_actions

    # ==================================================================
    # end of public methods

    # private methods:

    # def _send(): pass
    # def _register(): pass
    # def _create_listeners(): pass

    def _loop(self):
        # self._broker.run()
        self._broker.create_listeners(self.__loop)
        if self.__custom_event_loop:
            self.__loop.run_forever()
            self.__loop.close()

    # def _send(self, service: SERVICES, msg: str, *args):
    #     self._broker.send(service, msg, *args)

    def _configure(self):
        super()._configure()
        self._broker.register("event", self.__event_event)
        # super()._register("set_number_of_actions", self.__set_number_of_actions_event)
        self._broker.register(
            "set_current_capital_event", self.__set_current_capital_event
        )
        self._broker.register(
            "set_current_invested_event", self.__set_current_invested_event
        )
        self._broker.register("debug_breakpoint", self.__debug_breakpoint_event)
        self._broker.register("last_feed", self.__last_feed_event)

    # # override
    # def _asyncio_loop(self, loop: asyncio.AbstractEventLoop):
    #     self._broker.create_listeners(loop)

    # # override
    # def _handle_zmq_message(self, message):
    #     pass

    # COMMANDS

    async def __event_event(self, msg):
        await self.on_event(msg)

    # async def __set_number_of_actions_event(self, number: int):
    #     self.__number_of_actions = number

    async def __set_current_capital_event(self, number: int):
        self.__current_capital = number

    async def __set_current_invested_event(self, number: int):
        self.__current_invested = number

    async def __debug_breakpoint_event(self, breakpoint_params):
        await self._broker.send(
            SERVICES.python_backtester,
            "debug_breakpoint",
            DebugBreakpoint(**breakpoint_params),
        )

    async def __last_feed_event(self, msg):
        await self._broker.send(
            SERVICES.python_backtester, "last_feed", LastFeed(**msg)
        )



import importlib.util
from importlib.machinery import SourceFileLoader


def import_data_schema(path_to_strategy: str, strategy_file: str, data_class_name: str):
    module = SourceFileLoader(
        "data", path_to_strategy + "/" + strategy_file
    ).load_module()
    data_class = getattr(module, data_class_name)
    return data_class


def import_model_module(
    path_to_strategy: str, strategy_file: str, model_class_name: str
):
    module = SourceFileLoader(
        "model", path_to_strategy + "/" + strategy_file
    ).load_module()
    model_class = getattr(module, model_class_name)
    return model_class


def import_executor_module(
    path_to_strategy: str, strategy_file: str, executor_class_name: str
):
    module = SourceFileLoader(
        "executor", path_to_strategy + "/" + strategy_file
    ).load_module()
    executor_class = getattr(module, executor_class_name)
    return executor_class


def import_spec_module(path_to_module):
    spec = importlib.util.spec_from_file_location(path_to_module, path_to_module)
    module = importlib.util.module_from_spec(spec)
    return spec, module


def reload_spec_module(spec, module):
    spec.loader.exec_module(module)



import os
from backtesterrb30.libs.communication_broker.zmq_broker import ZMQ
from backtesterrb30.libs.utils.config_validator import validate_config
from backtesterrb30.libs.utils.module_loaders import import_data_schema
from backtesterrb30.libs.utils.service import Service


def run_service(microservice_name: str, service_class: Service):
    strategy_path = os.getenv("STRATEGY_PATH")
    sub_ports = [int(p) for p in os.getenv(microservice_name + "_subs").split(",")]
    pub_port = int(os.getenv(microservice_name + "_pubs"))
    strategy_file = os.getenv("STRATEGY_FILE")
    data_class_name = os.getenv("DATA_CLASS_NAME")
    backtest_state = os.getenv("backtest_state")
    from backtesterrb30.libs.interfaces.utils.config import Config

    config = {
        "name": microservice_name,
        "ip": "localhost",
        "sub": [{"topic": microservice_name, "port": p} for p in sub_ports],
        "pub": {"topic": "pub_" + str(pub_port), "port": pub_port},
        "backtest": True if backtest_state == "true" else False,
        "debug": True,
        "strategy_path": strategy_path,
    }
    data_class = import_data_schema(strategy_path, strategy_file, data_class_name)
    data_schema = validate_config(data_class.data)
    config = Config(**config)
    service: Service = service_class(config, data_schema)
    logger = service.get_logger()
    service.register_communication_broker(ZMQ(config, logger))
    service.run()



from enum import Enum
from typing import List, Optional

from pydantic import BaseModel

from backtesterrb30.libs.interfaces.utils.net_info import NetInfo


class BROKERS(Enum):
    zmq = ("zmq",)
    asyncio = "asyncio"


class Config(BaseModel):
    name: str
    ip: Optional[str]
    sub: Optional[List[NetInfo]]
    pub: Optional[NetInfo]
    backtest: bool
    debug: bool = False
    strategy_path: str



from enum import Enum


class SERVICES(Enum):
    historical_data_feeds = "historical_data_feeds"
    live_data_feeds = "live_data_feeds"
    python_backtester = "python_backtester"
    python_engine = "python_engine"
    python_executor = "python_executor"


SERVICES_ARRAY = [i for i in dir(SERVICES) if not i.startswith("__")]



import asyncio
import datetime
import enum
import json
import logging
import random
import re
import string

import pandas as pd
import requests
from websocket import create_connection

logger = logging.getLogger(__name__)


class Interval(enum.Enum):
    in_1_minute = "1"
    in_3_minute = "3"
    in_5_minute = "5"
    in_15_minute = "15"
    in_30_minute = "30"
    in_45_minute = "45"
    in_1_hour = "1H"
    in_2_hour = "2H"
    in_3_hour = "3H"
    in_4_hour = "4H"
    in_daily = "1D"
    in_weekly = "1W"
    in_monthly = "1M"


class TradingviewDownloader:
    sign_in_url = "https://www.tradingview.com/accounts/signin/"
    ws_headers = json.dumps({"Origin": "https://data.tradingview.com"})
    signin_headers = {"Referer": "https://www.tradingview.com"}
    ws_timeout = 5

    def __init__(
        self,
        username: str = None,
        password: str = None,
    ) -> None:
        """
        Create TvDatafeed object

        Args:
        ----
            username (str, optional): tradingview username. Defaults to None.
            password (str, optional): tradingview password. Defaults to None.

        """
        self.ws_debug = False

        self.token = self.auth(username, password)

        if self.token is None:
            self.token = "unauthorized_user_token"
            logger.warning(
                "you are using nologin method, data you access may be limited"
            )

        self.ws = None
        self.session = self.generate_session()
        self.chart_session = self.generate_chart_session()

    def auth(self, username, password):
        if username is None or password is None:
            token = None

        else:
            data = {"username": username, "password": password, "remember": "on"}
            try:
                response = requests.post(
                    url=self.sign_in_url, data=data, headers=self.signin_headers
                )
                token = response.json()["user"]["auth_token"]
            except Exception:
                logger.error("error while signin")
                token = None

        return token

    def create_connection(self):
        logging.debug("creating websocket connection")
        self.ws = create_connection(
            "wss://data.tradingview.com/socket.io/websocket",
            headers=self.ws_headers,
            timeout=self.ws_timeout,
        )

    @staticmethod
    def filter_raw_message(text):
        try:
            found = re.search('"m":"(.+?)",', text).group(1)
            found2 = re.search('"p":(.+?"}"])}', text).group(1)

            return found, found2
        except AttributeError:
            logger.error("error in filter_raw_message")

    @staticmethod
    def generate_session():
        stringLength = 12
        letters = string.ascii_lowercase
        random_string = "".join(random.choice(letters) for i in range(stringLength))
        return "qs_" + random_string

    @staticmethod
    def generate_chart_session():
        stringLength = 12
        letters = string.ascii_lowercase
        random_string = "".join(random.choice(letters) for i in range(stringLength))
        return "cs_" + random_string

    @staticmethod
    def prepend_header(st):
        return "~m~" + str(len(st)) + "~m~" + st

    @staticmethod
    def construct_message(func, param_list):
        return json.dumps({"m": func, "p": param_list}, separators=(",", ":"))

    def create_message(self, func, paramList):
        return self.prepend_header(self.construct_message(func, paramList))

    def send_message(self, func, args):
        m = self.create_message(func, args)
        if self.ws_debug:
            print(m)
        self.ws.send(m)

    @staticmethod
    def create_df(raw_data, symbol):
        try:
            out = re.search(r'"s":\[(.+?)\}\]', raw_data).group(1)
            x = out.split(',{"')
            data = []
            volume_data = True

            for xi in x:
                xi = re.split(r"\[|:|,|\]", xi)
                ts = int(float(xi[4]) * 1000)

                row = [ts]

                for i in range(5, 10):
                    # skip converting volume data if does not exists
                    if not volume_data and i == 9:
                        row.append(0.0)
                        continue
                    try:
                        row.append(float(xi[i]))

                    except ValueError:
                        volume_data = False
                        row.append(0.0)
                        logger.debug("no volume data")

                data.append(row)

            data = pd.DataFrame(
                data, columns=["timestamp", "open", "high", "low", "close", "volume"]
            )
            data.insert(0, "symbol", value=symbol)
            return data
        except AttributeError:
            logger.error("no data, please check the exchange and symbol")
            return pd.DataFrame()

    @staticmethod
    def format_symbol(symbol, exchange, contract: int = None):
        if ":" in symbol:
            pass
        elif contract is None:
            symbol = f"{exchange}:{symbol}"

        elif isinstance(contract, int):
            symbol = f"{exchange}:{symbol}{contract}!"

        else:
            raise ValueError("not a valid contract")

        return symbol

    async def get_hist(
        self,
        symbol: str,
        from_timestamp: int,
        exchange: str = "NSE",
        interval: Interval = Interval.in_daily,
        fut_contract: int = None,
        extended_session: bool = False,
    ) -> pd.DataFrame:
        """
        get historical data

        Args:
        ----
            symbol (str): symbol name
            exchange (str, optional): exchange, not required if symbol is in format EXCHANGE:SYMBOL. Defaults to None.
            interval (str, optional): chart interval. Defaults to 'D'.
            n_bars (int, optional): no of bars to download, max 5000. Defaults to 10.
            fut_contract (int, optional): None for cash, 1 for continuous current contract in front, 2 for continuous next contract in front . Defaults to None.
            extended_session (bool, optional): regular session if False, extended session if True, Defaults to False.

        Returns:
        -------
            pd.Dataframe: dataframe with sohlcv as columns

        """
        symbol = self.format_symbol(
            symbol=symbol, exchange=exchange, contract=fut_contract
        )

        interval = interval.value

        self.create_connection()

        self.send_message("set_auth_token", [self.token])
        self.send_message("chart_create_session", [self.chart_session, ""])
        self.send_message("quote_create_session", [self.session])
        self.send_message(
            "quote_set_fields",
            [
                self.session,
                "ch",
                "chp",
                "current_session",
                "description",
                "local_description",
                "language",
                "exchange",
                "fractional",
                "is_tradable",
                "lp",
                "lp_time",
                "minmov",
                "minmove2",
                "original_name",
                "pricescale",
                "pro_name",
                "short_name",
                "type",
                "update_mode",
                "volume",
                "currency_code",
                "rchp",
                "rtc",
            ],
        )

        # print(self.session)

        self.send_message(
            "quote_add_symbols", [self.session, symbol, {"flags": ["force_permission"]}]
        )
        self.send_message("quote_fast_symbols", [self.session, symbol])

        self.send_message(
            "resolve_symbol",
            [
                self.chart_session,
                "symbol_1",
                '={"symbol":"'
                + symbol
                + '","adjustment":"splits","session":'
                + ('"regular"' if not extended_session else '"extended"')
                + "}",
            ],
        )
        self.send_message(
            "create_series",
            [self.chart_session, "sds_1", "s1", "symbol_1", interval, 1000],
        )
        self.send_message("switch_timezone", [self.chart_session, "exchange"])

        raw_data = ""

        logger.debug(f"getting data for {symbol}...")
        while True:
            try:
                result = self.ws.recv()
                raw_data = raw_data + result + "\n"
            except Exception as e:
                logger.error(e)
                break

            if "series_completed" in result:
                break

        df = self.create_df(raw_data, symbol)

        list_dfs = []
        list_dfs.append(df)
        encountered_part_frames_number = 0
        encountered_zero_frames_number = 0
        while True:
            self.send_message(
                "request_more_data",
                [self.chart_session, "sds_1", 1000],
            )
            # result = tv.ws.recv()
            raw_data2 = ""
            while True:
                try:
                    result = self.ws.recv()
                    # print(result)
                    raw_data2 = raw_data2 + result + "\n"
                except Exception as e:
                    logger.error(e)
                    break

                if "series_completed" in result:
                    # print('completed 2')

                    break

                if result == "":
                    break
            # time.sleep(randint(1,3))

            df = self.create_df(raw_data2, symbol)
            if len(df) > 0:
                list_dfs.append(df)
            # print('len df', len(df))
            if len(df) < 0 and len(df) < 1000:
                if encountered_part_frames_number == 1:
                    print("second part freame. sleepin 20 s")
                    asyncio.sleep(20)
                encountered_part_frames_number += 1
            if len(df) == 0:
                if encountered_zero_frames_number == 1:
                    raise Exception(
                        "Tradingview data out of range."
                        + str(symbol)
                        + "\
                         "
                        + str(exchange)
                        + " "
                        + str(fut_contract)
                        + " Last avaliable date for interval "
                        + str(
                            interval
                            + " is "
                            + str(
                                datetime.datetime.utcfromtimestamp(
                                    list_dfs[-1].iloc[-1, 1] / 1000
                                )
                            )
                        )
                    )
                encountered_zero_frames_number += 1

            if len(df) > 0:
                print(df.iloc[-1, 1])
                if df.iloc[-1, 1] <= from_timestamp:
                    # print('smaller')
                    break

        # print(list_dfs)
        # reversed = list_dfs.reverse()
        # print(reversed)
        return pd.concat(list_dfs[::-1])


# if __name__ == "__main__":
#     logging.basicConfig(level=logging.DEBUG)
#     tv = TvDatafeed(
#     )
#     print(tv.get_hist("CRUDEOIL", "MCX", fut_contract=1))
#     # print(tv.get_hist("NIFTY", "NSE", fut_contract=1))
#     # print(
#     #     tv.get_hist(
#     #         "EICHERMOT",
#     #         "NSE",
#     #         interval=Interval.in_1_hour,
#     #         n_bars=500,
#     #         extended_session=False,
#     #     )
#     # )



from typing import List

from pydantic import BaseModel

from backtesterrb30.libs.interfaces.historical_data_feeds.instrument_file import (
    InstrumentFile,
)


class DataStartFiles(BaseModel):
    files: List[InstrumentFile]
    symbol: str
    source: str


class DataStart(BaseModel):
    file_names: List[DataStartFiles]
    start_time: float
